"""
This module contains utility functions for working with Polars DataFrames.
"""
import polars as pl

def rescale_column(df: pl.DataFrame,
                   column: str,
                   minimum: float | None = None,
                   maximum: float | None = None) -> pl.DataFrame:
    """
    Rescales a column to a range of [0, 1].

    Args:
        df (pl.DataFrame): A Polars DataFrame.
        column (str): The name of the column to rescale.
        minimum (float): 
            The minimum value of the column.
            If None, the minimum value of the
            column is used.
        maximum (float):
            The maximum value of the column.
            If None, the maximum value of the
            column is used.

    Returns:
        rescaled_df (pl.DataFrame):
            A Polars DataFrame with the column rescaled to [0, 1].
    """
    if minimum is None:
        minimum = df[column].min()
    if maximum is None:
        maximum = df[column].max()
    
    rescaled_df = df.with_columns([
        ((pl.col(column) - minimum) / (maximum - minimum)).alias(column)
    ])

    return rescaled_df

def normalize_column(df: pl.DataFrame,
                     column: str) -> pl.DataFrame:
    """
    Normalizes a column to have a mean of 0 and a standard deviation of 1.

    Args:
        df (pl.DataFrame): A Polars DataFrame.
        column (str): The name of the column to normalize.
    
    Returns:
        normalized_df (pl.DataFrame):
            A Polars DataFrame with the column normalized
    """
    mean = df[column].mean()
    std = df[column].std()

    normalized_df = df.with_columns([
        ((pl.col(column) - mean) / std).alias(column)
    ])

    return normalized_df

def filter_lexicon(lexicon: pl.DataFrame,
                   words: pl.Series,
                   word_column: str = "Word"
                   ) -> pl.DataFrame:
    """
    Filters a lexicon to only include the words in a list.

    Args:
        lexicon (pl.DataFrame): A Polars DataFrame containing the lexicon.
        words (pl.Series): A Polars Series containing the words to include.
        word_column (str):
            The name of the column containing the words in the lexicon.

    Returns:
        filtered_lexicon (pl.DataFrame):
            A Polars DataFrame containing only the words in the list.
    """
    return lexicon.filter(pl.col(word_column).is_in(words))

def upos_to_wn(upos_tag: str) -> str:
    """
    Converts a Universal POS tag to a (Senti)WordNet POS tag.

    Args:
        upos_tag (str): A Universal POS tag.

    Returns:
        wn_tag (str): A WordNet POS tag.
    """
    if upos_tag in {"NOUN", "PROPN"}:
        return "n"
    elif upos_tag in {"VERB", "AUX"}:
        return "v"
    elif upos_tag in {"ADV"}:
        return "r"
    elif upos_tag in {"ADJ"}:
        return "a"
    else:
        return None




"""
Generic functions for custom lexicon feature extraction. These functions
allow the user to extract features from a custom lexicon. The lexicon
should be a DataFrame with with the tokens or lemmas in one column and
the features (e.g. concreteness; only if thresholds are to be applied) in
another column.

Functions:
    get_n_custom: Get the number of occurences of words in a custom lexicon
        in the text.
    get_occurs_custom: Binary feature indicating if any word from a custom
        lexicon occurs in the text.
    get_n_custom_low: Get the number of occurences of words with a low
        given feature (e.g. concreteness) from a custom lexicon in the
        text.
    get_n_custom_high: Get the number of occurences of words with a high
        given feature (e.g. concreteness) from a custom lexicon in the
        text.
"""

import polars as pl

from .preprocess import (
    get_lemmas,
    get_tokens,
)
from .util import (
    filter_lexicon,
)

def get_n_custom(data: pl.DataFrame,
                 lexicon: pl.DataFrame,
                 feature_name: str = 'n_custom',
                 word_column: str = 'word',
                 measurement_level: str = 'tokens',
                 ) -> pl.DataFrame:
    """
    Get the number of occurences of words in a custom lexicon in the text.

    Args:
        data (pl.DataFrame):
            The data to process. Should contain the preprocessed column
            'nlp'.
        lexicon (pl.DataFrame): The lexicon to use.
        feature_name (str): The name of the feature to create.
        word_column (str): The column the words occur in in the lexicon.
        measurement_level (str):
            The measurement level of the lexicon; either 'tokens' or
            'lemmas'.

    Returns:
        data (pl.DataFrame): 
            The data with the new feature.
    """
    if measurement_level == 'token' and 'tokens' not in data.columns:
        data = get_tokens(data)
    elif measurement_level == 'lemmas' and 'lemmas' not in data.columns:
        data = get_lemmas(data)

    data = data.with_columns([
        pl.col(measurement_level).map_elements(
            lambda x: len(filter_lexicon(lexicon=lexicon,
                                         words=x,
                                         word_column=word_column,)),
            return_dtype=pl.UInt32).alias(feature_name)
    ])

    return data
    
def get_occurs_custom(data: pl.DataFrame,
                      lexicon: pl.DataFrame,
                      feature_name: str = 'occurs_custom',
                      word_column: str = 'word',
                      measurement_level: str = 'tokens',
                      ) -> pl.DataFrame:
    """
    Binary feature indicating if any word from a custom lexicon occurs in
    the text.

    Args:
        data (pl.DataFrame):
            The data to process. Should contain the preprocessed column
            'nlp'.
        lexicon (pl.DataFrame): The lexicon to use.
        feature_name (str): The name of the feature to create.
        word_column (str): The column the words occur in in the lexicon.

    Returns:
        data (pl.DataFrame): 
            The data with the new feature.
    """
    if measurement_level == 'token' and 'tokens' not in data.columns:
        data = get_tokens(data)
    elif measurement_level == 'lemmas' and 'lemmas' not in data.columns:
        data = get_lemmas(data)

    data = data.with_columns([
        (~pl.col(measurement_level).map_elements(
            lambda x: filter_lexicon(lexicon=lexicon,
                                     words=x,
                                     word_column=word_column,). \
            is_empty(),
            return_dtype=pl.Boolean)).alias(feature_name)
    ])

    return data

def get_n_custom_low(data: pl.DataFrame,
                     lexicon: pl.DataFrame,
                     threshold: int,
                     feature_name: str = 'n_custom_low',
                     word_column: str = 'word',
                     feature_column: str = 'feature',
                     measurement_level: str = 'tokens',
                     ) -> pl.DataFrame:
    """
    Get the number of occurences of words in a custom lexicon in the text
    with a given feature (e.g. concreteness) below a threshold.

    Args:
        data (pl.DataFrame):
            The data to process. Should contain the preprocessed column
            'nlp'.
        lexicon (pl.DataFrame): The lexicon to use.
        feature_name (str): The name of the feature to create.
        threshold (int): The threshold to use.
        word_column (str): The column the words occur in in the lexicon.
        feature_column (str):
            The column the feature occurs in in the lexicon.
        measurement_level (str):
            The measurement level of the lexicon; either 'tokens' or
            'lemmas'.

    Returns:
        data (pl.DataFrame): 
            The data with the new feature.
    """
    if measurement_level == 'token' and 'tokens' not in data.columns:
        data = get_tokens(data)
    elif measurement_level == 'lemmas' and 'lemmas' not in data.columns:
        data = get_lemmas(data)

    data = data.with_columns([
        pl.col(measurement_level).map_elements(
            lambda x: len(filter_lexicon(lexicon=lexicon,
                                         words=x,
                                         word_column=word_column,). \
            filter(pl.col(feature_column) < threshold)),
            return_dtype=pl.UInt32).alias(feature_name)
    ])

    return data

def get_n_custom_high(data: pl.DataFrame,
                      lexicon: pl.DataFrame,
                      threshold: int,
                      feature_name: str = 'n_custom_high',
                      word_column: str = 'word',
                      feature_column: str = 'feature',
                      measurement_level: str = 'tokens',
                      ) -> pl.DataFrame:
    """
    Get the number of occurences of words in a custom lexicon in the text
    with a given feature (e.g. concreteness) above a threshold.

    Args:
        data (pl.DataFrame):
            The data to process. Should contain the preprocessed column
            'nlp'.
        lexicon (pl.DataFrame): The lexicon to use.
        threshold (int): The threshold to use.
        feature_name (str): The name of the feature to create.
        word_column (str): The column the words occur in in the lexicon.
        feature_column (str):
            The column the feature occurs in in the lexicon.
        measurement_level (str):
            The measurement level of the lexicon; either 'tokens' or
            'lemmas'.

    Returns:
        data (pl.DataFrame): 
            The data with the new feature.
    """
    if measurement_level == 'token' and 'tokens' not in data.columns:
        data = get_tokens(data)
    elif measurement_level == 'lemmas' and 'lemmas' not in data.columns:
        data = get_lemmas(data)

    data = data.with_columns([
        pl.col(measurement_level).map_elements(
            lambda x: len(filter_lexicon(lexicon=lexicon,
                                     words=x,
                                     word_column=word_column,). \
            filter(pl.col(feature_column) > threshold)),
            return_dtype=pl.UInt32).alias(feature_name)
    ])

    return data


def get_avg_custom(data: pl.DataFrame,
                   lexicon: pl.DataFrame,
                   feature_name: str = 'avg_custom',
                   word_column: str = 'word',
                   feature_column: str = 'feature',
                   measurement_level: str = 'tokens',
                   ) -> pl.DataFrame:
    """
    Get the average feature value of words in a custom lexicon in the text.

    Args:
        data (pl.DataFrame):
            The data to process. Should contain the preprocessed column
            'nlp'.
        lexicon (pl.DataFrame): The lexicon to use.
        feature_name (str): The name of the feature to create.
        word_column (str): The column the words occur in in the lexicon.
        feature_column (str):
            The column the feature occurs in in the lexicon.
        measurement_level (str):
            The measurement level of the lexicon; either 'tokens' or
            'lemmas'.

    Returns:
        data (pl.DataFrame): 
            The data with the new feature.
    """
    if measurement_level == 'token' and 'tokens' not in data.columns:
        data = get_tokens(data)
    elif measurement_level == 'lemmas' and 'lemmas' not in data.columns:
        data = get_lemmas(data)

    data = data.with_columns([
        pl.col(measurement_level).map_elements(
            lambda x: filter_lexicon(lexicon=lexicon,
                                     words=x,
                                     word_column=word_column,). \
            select(pl.col(feature_column)).mean().item()
            ).alias(feature_name)
    ])

    return data




"""
This module contains functions to calculate the VAD dimensions, Plutchik
emotions, the sentiment and emotion intensity of text data.

If you are using this functionality, please cite the original authors of
the resources.

The lexicons used in this module are:

- NRC VAD Lexicon:
    Mohammad, S. M., & Turney, P. D. (2013).
    Crowdsourcing a Word-Emotion Association Lexicon.
    Computational Intelligence, 29(3), 436-465.
- NRC Emotion Intensity Lexicon:
    Mohammad, S. M. (2018).
    Obtaining reliable human ratings of valence, arousal, and dominance
    for 20,000 English words.
    In Proceedings of the 56th Annual Meeting of the Association for
    Computational Linguistics (Volume 1: Long Papers) (pp. 174-184).
- NRC Sentiment Lexicon:
    Mohammad, S.M. (2018).
    Word Affect Intensities.
    In Proceedings of the 11th Edition of the Language Resources and 
    Evaluation Conference (LREC-2018), May 2018, Miyazaki, Japan.

The VAD lexicon contains three dimensions: Valence, Arousal, and Dominance.

The emotion intensity lexicon contains the intensity of eight Plutchik
emotions: Anger, Anticipation, Disgust, Fear, Joy, Sadness, Surprise,
and Trust.

The sentiment lexicon contains the sentiment of words: Positive and
Negative.

The following functions are implemented in this module:

- VAD Dimensions:

    - load_vad_lexicon: 
        Loads the VAD lexicon as a polars DataFrame.
    - get_avg_valence:
        Calculates the average valence of the text.
    - get_avg_arousal:
        Calculates the average arousal of the text.
    - get_avg_dominance:
        Calculates the average dominance of the text.
    - get_n_low_valence
        Calculates the number of words with valence lower than the
        threshold.
    - get_n_high_valence:
        Calculates the number of words with valence higher than the
        threshold.
    - get_n_low_arousal:
        Calculates the number of words with arousal lower than the
        threshold.
    - get_n_high_arousal:
        Calculates the number of words with arousal higher than the
        threshold.
    - get_n_low_dominance:
        Calculates the number of words with dominance lower than the
        threshold.
    - get_n_high_dominance:
        Calculates the number of words with dominance higher than the
        threshold.

- Emotion Intensity:

    - load_intensity_lexicon:
        Loads the emotion intensity lexicon as a polars DataFrame.
    - get_avg_emotion_intensity:
        Calculates the average emotion intensity of the text.
    - get_n_low_intensity:
        Calculates the number of words with emotion intensity lower than
        the threshold.
    - get_n_high_intensity:
        Calculates the number of words with emotion intensity higher than
        the threshold.
    
- Sentiment Analysis:

    - load_sentiment_nrc_lexicon:
        Loads the sentiment NRC lexicon as a polars DataFrame.
    - get_n_positive_sentiment:
        Calculates the number of words with positive sentiment.
    - get_n_negative_sentiment:
        Calculates the number of words with negative sentiment.
"""
import polars as pl

from .preprocess import (
    get_lemmas,
)
from .resources import (
    RESOURCE_MAP,
    LANGUAGES_NRC,
)
from .schemas import (
    VAD_SCHEMA_NRC,
    VAD_SCHEMA_NRC_MULTILINGUAL,
    INTENSITY_SCHEMA,
    INTENSITY_SCHEMA_MULTILINGUAL,
    SENTIMENT_NRC_SCHEMA,
    SENTIMENT_NRC_SCHEMA_MULTILINGUAL,
)
from .surface import (
    get_num_tokens,
)
from .util import (
    filter_lexicon,
)

EMOTIONS = [
    "anger",
    "anticipation",
    "disgust",
    "fear",
    "joy",
    "sadness",
    "surprise",
    "trust"
]

VAD_NRC_PATH = RESOURCE_MAP["vad_nrc"]["filepath"]
INTENSITY_PATH = RESOURCE_MAP["intensity_nrc"]["filepath"]
SENTIWORDNET_PATH = RESOURCE_MAP["sentiwordnet"]["filepath"]
SENTIMENT_NRC_PATH = RESOURCE_MAP["sentiment_nrc"]["filepath"]

# --------------------------------------------------------------------- #
#                           VAD dimensions                              #
# --------------------------------------------------------------------- #

def load_vad_lexicon(path: str = VAD_NRC_PATH,
                     language: str = "en",
                     has_header: bool = True,
                     separator: str = "\t",
                     **kwargs: dict[str, str],
                     ) -> pl.DataFrame:
    """
    Loads the VAD lexicon as a polars DataFrame.

    Args:
        path (str): The path to the VAD lexicon.
        schema (dict): The schema for the VAD lexicon.
        has_header (bool): Whether the lexicon has a header.
        separator (str): The separator used in the lexicon.

    Returns:
        vad_lexicon (pl.DataFrame):
            The VAD lexicon as a polars DataFrame.
    """
    if language != "en":
        schema = VAD_SCHEMA_NRC_MULTILINGUAL
    else:
        schema = VAD_SCHEMA_NRC

    vad_lexicon = pl.read_csv(path,
                              has_header=has_header,
                              schema=schema,
                              separator=separator)
    if "V.Mean" in vad_lexicon.columns: # rename columns for consistency
        vad_lexicon = vad_lexicon.rename(
            {"V.Mean": "valence",
             "A.Mean": "arousal",
             "D.Mean": "dominance"}
        )
    elif "Valence" in vad_lexicon.columns:
        vad_lexicon = vad_lexicon.rename(
            {"Valence": "valence",
             "Arousal": "arousal",
             "Dominance": "dominance"}
        )
    return vad_lexicon

def get_avg_valence(data: pl.DataFrame,
                    lexicon: pl.DataFrame,
                    backbone: str = "spacy",
                    language: str = "en",
                    nan_value: float = 0.0,
                    **kwargs: dict[str, str],
                    ) -> pl.DataFrame:
    """
    Calculates the average valence of the text. Only takes into account
    words in the text that are present in the VAD lexicon.

    The valence of the text is calculated as the mean of the valence
    values of the words in the text.

    NaN/Null values are filled with 0 as the valence is in the range [0,1]
    and 0 is the neutral value in the NRC-VAD lexicon.

    Args:
        data (pl.DataFrame):
            The preprocessed input data. Contains the "nlp" column
            produced by the NLP backbone.
        lexicon (pl.DataFrame): The VAD lexicon.
        backbone (str): The NLP backbone to use.
        nan_value (float): The value to use for NaNs. Defaults to 0.0.

    Returns:
        data (pl.DataFrame):
            The input data with the average valence columns,
            named "avg_valence".
    """
    if "lemmas" not in data.columns:
        data = get_lemmas(data, backbone=backbone)

    if language != "en":
        word_column = LANGUAGES_NRC[language]
    else:
        word_column = "word"
    
    data = data.with_columns(
        pl.col("lemmas").map_elements(
            lambda x: filter_lexicon(lexicon=lexicon,
                                     words=x,
                                     word_column=word_column). \
               select("valence").mean().item(),
               return_dtype=pl.Float64
        ).fill_nan(nan_value).fill_null(nan_value).alias("avg_valence")
        # convention to fill NaNs with 0 as valence is in [0,1]
        # and 0 is the neutral value for the NRC-VAD lexicon
    )
    
    return data

def get_avg_arousal(data: pl.DataFrame,
                    lexicon: pl.DataFrame,
                    backbone: str = "spacy",
                    nan_value: float = 0.0,
                    language: str = "en",
                    **kwargs: dict[str, str],
                    ) -> pl.DataFrame:
    """
    Calculates the average arousal of the text. Only takes into account
    words in the text that are present in the VAD lexicon.
    
    The arousal of the text is calculated as the mean of the arousal
    values of the words in the text.

    NaN/Null values are filled with 0 as the arousal is in the range [0,1]
    and 0 is the neutral value in the NRC-VAD lexicon.

    Args:
        data (pl.DataFrame):
            The preprocessed input data. Contains the "nlp" column
            produced by the NLP backbone.
        lexicon (pl.DataFrame): The VAD lexicon.
        backbone (str): The NLP backbone to use.
        nan_value (float): The value to use for NaNs.
                         Defaults to 0.0.
    
    Returns:
        data (pl.DataFrame):
            The input data with the average arousal columns,
            named "avg_arousal".
    """
    if "lemmas" not in data.columns:
        data = get_lemmas(data, backbone=backbone)

    if language != "en":
        word_column = LANGUAGES_NRC[language]
    else:
        word_column = "word"
    
    data = data.with_columns(
        pl.col("lemmas").map_elements(
             lambda x: filter_lexicon(lexicon=lexicon,
                                      words=x,
                                      word_column=word_column). \
               select("arousal").mean().item(),
               return_dtype=pl.Float64
        ).fill_nan(nan_value).fill_null(nan_value).alias("avg_arousal")
        # convention to fill NaNs with 0 as arousal is in [0,1]
        # and 0 is the neutral value for the NRC-VAD lexicon
    )
    
    return data

def get_avg_dominance(data: pl.DataFrame,
                      lexicon: pl.DataFrame,
                      backbone: str = "spacy",
                      nan_value: float = 0.0,
                      language: str = "en",
                      **kwargs: dict[str, str],
                      ) -> pl.DataFrame:
    """
    Calculates the average dominance of the text. Only takes into account
    words in the text that are present in the VAD lexicon.

    The dominance of the text is calculated as the mean of the dominance
    values of the words in the text.

    NaN/Null values are filled with 0 as the dominance is in the range
    [0,1] and 0 is the neutral value in the NRC-VAD lexicon.

    Args:
        data (pl.DataFrame):
            The preprocessed input data. Contains the "nlp" column
            produced by the NLP backbone.
        lexicon (pl.DataFrame): The VAD lexicon.
        backbone (str): The NLP backbone to use.
        nan_value (float): The value to use for NaNs.
                         Defaults to 0.0.

    Returns:
        data (pl.DataFrame):
            The input data with the average dominance columns,
            named "avg_dominance".
    """
    if "lemmas" not in data.columns:
        data = get_lemmas(data, backbone=backbone)

    if language != "en":
        word_column = LANGUAGES_NRC[language]
    else:
        word_column = "word"
    
    data = data.with_columns(
         pl.col("lemmas").map_elements(
          lambda x: filter_lexicon(lexicon=lexicon,
                                   words=x,
                                   word_column=word_column). \
                select("dominance").mean().item(),
                return_dtype=pl.Float64
         ).fill_nan(nan_value).fill_null(nan_value).alias("avg_dominance")
         # convention to fill NaNs with 0 as dominance is in [0,1]
         # and 0 is the neutral value for the NRC-VAD lexicon
    )

    return data

def get_n_low_valence(data: pl.DataFrame,
                      lexicon: pl.DataFrame,
                      backbone: str = "spacy",
                      threshold: float = 0.33,
                      nan_value: float = 0.0,
                      language: str = "en",
                      **kwargs: dict[str, str],
                      ) -> pl.DataFrame:
    """
    Calculates the number of words with valence lower than the threshold.

    Args:
        data (pl.DataFrame):
            The preprocessed input data. Contains the
            "nlp" column produced by the NLP backbone.
        lexicon (pl.DataFrame): The VAD lexicon.
        backbone (str): The NLP backbone to use.
        threshold (float): The threshold for low valence.
                            Defaults to 0.33.
        nan_value (float): The value to use for NaNs.
                            Defaults to 0.0.

    Returns:
        data (pl.DataFrame):
            The input data with the low valence count column,
            named "n_low_valence".
    """
    if "lemmas" not in data.columns:
        data = get_lemmas(data, backbone=backbone)

    if language != "en":
        word_column = LANGUAGES_NRC[language]
    else:
        word_column = "word"
    
    data = data.with_columns(
        pl.col("lemmas").map_elements(
             lambda x: filter_lexicon(lexicon=lexicon,
                                      words=x,
                                      word_column=word_column). \
               select("valence").filter(
                   pl.col("valence") < threshold).shape[0],
               return_dtype=pl.UInt32
        ).fill_nan(nan_value).fill_null(nan_value).alias("n_low_valence")
    )
    
    return data

def get_n_high_valence(data: pl.DataFrame,
                       lexicon: pl.DataFrame,
                       backbone: str = "spacy",
                       threshold: float = 0.66,
                       nan_value: float = 0.0,
                       language: str = "en",
                       **kwargs: dict[str, str],
                       ) -> pl.DataFrame:
    """
    Calculates the number of words with valence higher than the threshold.

    Args:
        data (pl.DataFrame):
            The preprocessed input data. Contains the "nlp" column
            produced by the NLP backbone.
        lexicon (pl.DataFrame): The VAD lexicon.
        backbone (str): The NLP backbone to use.
        threshold (float): The threshold for high valence.
                         Defaults to 0.66.
        nan_value (float): The value to use for NaNs.
                         Defaults to 0.0.

    Returns:
        data (pl.DataFrame):
            The input data with the high valence count column,
            named "n_high_valence".
    """
    if "lemmas" not in data.columns:
        data = get_lemmas(data, backbone=backbone)

    if language != "en":
        word_column = LANGUAGES_NRC[language]
    else:
        word_column = "word"
    
    data = data.with_columns(
        pl.col("lemmas").map_elements(
            lambda x: filter_lexicon(lexicon=lexicon,
                                      words=x,
                                      word_column=word_column). \
               select("valence").filter(
                   pl.col("valence") > threshold).shape[0],
               return_dtype=pl.UInt32
        ).fill_nan(nan_value).fill_null(nan_value).alias("n_high_valence")
    )
    
    return data

def get_n_low_arousal(data: pl.DataFrame,
                      lexicon: pl.DataFrame,
                      backbone: str = "spacy",
                      threshold: float = 0.33,
                      nan_value: float = 0.0,
                      language: str = "en",
                      **kwargs: dict[str, str],
                      ) -> pl.DataFrame:
    """
    Calculates the number of words with arousal lower than the threshold.

    Args:
        data (pl.DataFrame): 
            The preprocessed input data. Contains the
            "nlp" column produced by the NLP backbone.
        lexicon (pl.DataFrame): The VAD lexicon.
        backbone (str): The NLP backbone to use.
        threshold (float): The threshold for low arousal.
                         Defaults to 0.33.
        nan_value (float): The value to use for NaNs.
                         Defaults to 0.0.

    Returns:
        data (pl.DataFrame):
            The input data with the low arousal count column, named
            "n_low_arousal".
    """
    if "lemmas" not in data.columns:
        data = get_lemmas(data, backbone=backbone)
        
    if language != "en":
        word_column = LANGUAGES_NRC[language]
    else:
        word_column = "word"
    
    data = data.with_columns(
        pl.col("lemmas").map_elements(
             lambda x: filter_lexicon(lexicon=lexicon,
                                      words=x,
                                      word_column=word_column). \
               select("arousal").filter(
                   pl.col("arousal") < threshold).shape[0],
               return_dtype=pl.UInt32
        ).fill_nan(nan_value).fill_null(nan_value).alias("n_low_arousal")
    )
    
    return data

def get_n_high_arousal(data: pl.DataFrame,
                       lexicon: pl.DataFrame,
                       backbone: str = "spacy",
                       threshold: float = 0.66,
                       nan_value: float = 0.0,
                       language: str = "en",
                       **kwargs: dict[str, str],
                       ) -> pl.DataFrame:
    """
    Calculates the number of words with arousal higher than the threshold.

    Args:
        data (pl.DataFrame):
            The preprocessed input data. Contains the
            "nlp" column produced by the NLP backbone.
        lexicon (pl.DataFrame): The VAD lexicon.
        backbone (str): The NLP backbone to use.
        threshold (float): The threshold for high arousal.
                         Defaults to 0.66.
        nan_value (float): The value to use for NaNs.
                            Defaults to 0.0.

    Returns:
        data (pl.DataFrame):
            The input data with the high arousal count column. The
            column name is "n_high_arousal".
    """
    if "lemmas" not in data.columns:
        data = get_lemmas(data, backbone=backbone)
        
    if language != "en":
        word_column = LANGUAGES_NRC[language]
    else:
        word_column = "word"
    
    data = data.with_columns(
        pl.col("lemmas").map_elements(
             lambda x: filter_lexicon(lexicon=lexicon,
                                      words=x,
                                      word_column=word_column). \
               select("arousal").filter(
                   pl.col("arousal") > threshold).shape[0],
               return_dtype=pl.UInt32
        ).fill_nan(nan_value).fill_null(nan_value).alias("n_high_arousal")
    )
    
    return data

def get_n_low_dominance(data: pl.DataFrame,
                        lexicon: pl.DataFrame,
                        backbone: str = "spacy",
                        threshold: float = 0.33,
                        nan_value: float = 0.0,
                        language: str = "en",
                        **kwargs: dict[str, str],
                        ) -> pl.DataFrame:
    """
    Calculates the number of words with dominance lower than the threshold.

    Args:
        data (pl.DataFrame): 
            The preprocessed input data. Contains the
            "nlp" column produced by the NLP backbone.
        lexicon (pl.DataFrame): The VAD lexicon.
        backbone (str): The NLP backbone to use.
        threshold (float): The threshold for low dominance.
                         Defaults to 0.33.
        nan_value (float): The value to use for NaNs.
                            Defaults to 0.0.

    Returns:
        data (pl.DataFrame):
            The input data with the low dominance count column. The
            column name is "n_low_dominance".
    """
    if "lemmas" not in data.columns:
        data = get_lemmas(data, backbone=backbone)
        
    if language != "en":
        word_column = LANGUAGES_NRC[language]
    else:
        word_column = "word"
    
    data = data.with_columns(
        pl.col("lemmas").map_elements(
             lambda x: filter_lexicon(lexicon=lexicon,
                                      words=x,
                                      word_column=word_column). \
               select("dominance").filter(
                   pl.col("dominance") < threshold).shape[0],
               return_dtype=pl.UInt32
        ).fill_nan(nan_value).fill_null(nan_value).alias("n_low_dominance")
    )
    
    return data

def get_n_high_dominance(data: pl.DataFrame,
                         lexicon: pl.DataFrame,
                         backbone: str = "spacy",
                         threshold: float = 0.66,
                         nan_value: float = 0.0,
                         language: str = "en",
                         **kwargs: dict[str, str],
                         ) -> pl.DataFrame:
    """
    Calculates the number of words with dominance higher than the
    threshold.

    Args:
        data (pl.DataFrame):
            The preprocessed input data. Contains the
            "nlp" column produced by the NLP backbone.
        lexicon (pl.DataFrame): The VAD lexicon.
        backbone (str): The NLP backbone to use.
        threshold (float): The threshold for high dominance.
                            Defaults to 0.66.
        nan_value (float): The value to use for NaNs.
                            Defaults to 0.0.

    Returns:
        data (pl.DataFrame):
            The input data with the high dominance count column. The
            column name is "n_high_dominance".
    """
    if "lemmas" not in data.columns:
        data = get_lemmas(data, backbone=backbone)
    
    if language != "en":
        word_column = LANGUAGES_NRC[language]
    else:
        word_column = "word"
    
    data = data.with_columns(
        pl.col("lemmas").map_elements(
             lambda x: filter_lexicon(lexicon=lexicon,
                                      words=x,
                                      word_column=word_column). \
               select("dominance").filter(
                   pl.col("dominance") > threshold).shape[0],
               return_dtype=pl.UInt32
        ).fill_nan(nan_value).fill_null(nan_value).alias("n_high_dominance")
    )
    
    return data

# --------------------------------------------------------------------- #
#                           EMOTION INTENSITY                           #
# --------------------------------------------------------------------- #

def load_intensity_lexicon(path: str = INTENSITY_PATH,
                           language: str = "en",
                           has_header: bool = True,
                           separator: str = "\t",
                           **kwargs: dict[str, str],
                           ) -> pl.DataFrame:
    """
    Loads the intensity lexicon as a polars DataFrame.

    Args:
        path (str): The path to the intensity lexicon.
        schema (dict): The schema for the intensity lexicon.
        has_header (bool): Whether the lexicon has a header.

    Returns:
        intensity_lexicon (pl.DataFrame):
            The intensity lexicon as a polars DataFrame.
    """
    if language != "en":
        schema = INTENSITY_SCHEMA_MULTILINGUAL
    else:
        schema = INTENSITY_SCHEMA

    intensity_lexicon = pl.read_csv(path,
                                    has_header=has_header,
                                    schema=schema,
                                    separator=separator)
    # rename columns for consistency
    if "Emotion-Intensity-Score" in intensity_lexicon.columns:
        intensity_lexicon = intensity_lexicon.rename(
            {"Emotion-Intensity-Score": "emotion_intensity",
             "Emotion": "emotion"}
        )
    return intensity_lexicon

def filter_intensity_lexicon(lexicon: pl.DataFrame,
                             words: list,
                             emotion: str,
                             word_column: str = "word",
                             **kwargs: dict[str, str],
                             ) -> pl.DataFrame:
    """
    Filters the intensity lexicon for the given words and emotions.

    Args:
        lexicon (pl.DataFrame): The emotion intensity lexicon.
        words (list): The list of words to filter.
        emotion (str): The emotion to filter for.

    Returns:
        filtered_intensity_lexicon (pl.DataFrame):
            The filtered emotion intensity lexicon.
    """
    filtered_intensity_lexicon = lexicon.filter(
        (pl.col(word_column).is_in(words)) &
        (pl.col("emotion") == emotion)
    )
    
    return filtered_intensity_lexicon

def get_avg_emotion_intensity(data: pl.DataFrame,
                              lexicon: pl.DataFrame,
                              backbone: str = "spacy",
                              emotions: list = EMOTIONS,
                              nan_value: float = 0.0,
                              language: str = "en",
                              **kwargs: dict[str, str],
                              ) -> pl.DataFrame:
    """
    Calculates the average emotion intensity of the text. Only takes into
    account words in the text that are present in the emotion intensity
    lexicon.

    The average emotion intensity is calculated as the mean of the emotion
    intensity values of the words in the text.

    NaN/Null values are filled with 0 as the emotion intensity is in the
    range [0,1] and 0 is the neutral value in the NRC emotion intensity
    lexicon.

    Args:
        data (pl.DataFrame):
            The preprocessed input data. Contains the
            "nlp" column produced by the NLP backbone.
        lexicon (pl.DataFrame): The emotion intensity lexicon.
        backbone (str): The NLP backbone to use.
        emotions (list): The list of emotions to consider.
                       Defaults to the Plutchik emotions.
        nan_value (float): The value to use for NaNs.
                            Defaults to 0.0.

    Returns:
        data (pl.DataFrame):
            The input data with the average emotion intensity columns
    """
    if "lemmas" not in data.columns:
        data = get_lemmas(data, backbone=backbone)
        
    if language != "en":
        word_column = LANGUAGES_NRC[language]
    else:
        word_column = "word"
    
    for emotion in emotions:
        data = data.with_columns(
            pl.col("lemmas").map_elements(
                lambda x: filter_intensity_lexicon(lexicon,
                    x, emotion, word_column=word_column). \
                    select("emotion_intensity").mean().item(),
                return_dtype=pl.Float64
            ).fill_nan(nan_value).fill_null(nan_value). \
                alias(f"avg_intensity_{emotion}")
        )

    return data

def get_n_low_intensity(data: pl.DataFrame,
                        lexicon: pl.DataFrame,
                        backbone: str = "spacy",
                        emotions: list = EMOTIONS,
                        threshold: float = 0.33,
                        nan_value: float = 0.0,
                        language: str = "en",
                        **kwargs: dict[str, str],
                        ) -> pl.DataFrame:
    """
    Calculates the number of words with emotion intensity lower than the
    threshold.

    Args:
        data (pl.DataFrame): The preprocessed input data. Contains the
                           "nlp" column produced by the NLP backbone.
        lexicon (pl.DataFrame): The emotion intensity lexicon.
        backbone (str): The NLP backbone to use.
        emotions (list): The list of emotions to consider.
        threshold (float): The threshold for low intensity.
                         Defaults to 0.33.
        nan_value (float): The value to use for NaNs.
                         Defaults to 0.0.

    Returns:
        data (pl.DataFrame):
            The input data with the low intensity count column for each
            emotion. The column names are in the format
            "n_low_intensity_{emotion}".
    """
    if "lemmas" not in data.columns:
        data = get_lemmas(data, backbone=backbone)
        
    if language != "en":
        word_column = LANGUAGES_NRC[language]
    else:
        word_column = "word"
    
    for emotion in emotions:
        data = data.with_columns(
            pl.col("lemmas").map_elements(
                lambda x: filter_intensity_lexicon(
                    lexicon, x, emotion, word_column=word_column). \
                    select("emotion_intensity").filter(
                        pl.col("emotion_intensity") < threshold).shape[0],
                return_dtype=pl.UInt32
            ).fill_null(nan_value).fill_nan(nan_value). \
                alias(f"n_low_intensity_{emotion}")
        )

    return data

def get_n_high_intensity(data: pl.DataFrame,
                         lexicon: pl.DataFrame,
                         backbone: str = "spacy",
                         emotions: list = EMOTIONS,
                         threshold: float = 0.66,
                         nan_value: float = 0.0,
                         language: str = "en",
                         **kwargs: dict[str, str],
                         ) -> pl.DataFrame:
    """
    Calculates the number of words with emotion intensity higher than the
    threshold.

    Args:
        data (pl.DataFrame):
            The preprocessed input data. Contains the
            "nlp" column produced by the NLP backbone.
        lexicon (pl.DataFrame): The emotion intensity lexicon.
        backbone (str): The NLP backbone to use.
        emotions (list): The list of emotions to consider.
        threshold (float): The threshold for high intensity.
                            Defaults to 0.66.
        nan_value (float): The value to use for NaNs.
                            Defaults to 0.0.

    Returns:
        data (pl.DataFrame):
            The input data with the high intensity count column for each
            emotion. The column names are in the format
            "n_high_intensity_{emotion}".
    """
    if "lemmas" not in data.columns:
        data = get_lemmas(data, backbone=backbone)
    
    if language != "en":
        word_column = LANGUAGES_NRC[language]
    else:
        word_column = "word"
    
    for emotion in emotions:
        data = data.with_columns(
            pl.col("lemmas").map_elements(
                lambda x: filter_intensity_lexicon(
                    lexicon, x, emotion, word_column=word_column). \
                    select("emotion_intensity").filter(
                        pl.col("emotion_intensity") > threshold).shape[0],
                return_dtype=pl.UInt32
            ).fill_nan(nan_value).fill_null(nan_value). \
                alias(f"n_high_intensity_{emotion}")
        )

    return data

# --------------------------------------------------------------------- #
#                           SENTIMENT ANALYSIS                          #
# --------------------------------------------------------------------- #

# ---------------------------- SentiWordNet --------------------------- #
# NOTE: SentiWordNet is not used in the current implementation

# def load_sentiwordnet(path: str = SENTIWORDNET_PATH,
#                       schema: dict = SENTIWORDNET_SCHEMA,
#                       has_header: bool = False,
#                       separator: str = "\t",
#                       **kwargs: dict[str, str],
#                       ) -> pl.DataFrame:
#     """
#     Loads the SentiWordNet lexicon as a polars DataFrame.

#     Args:
#         path (str): The path to the SentiWordNet lexicon.
#         schema (dict): The schema for the SentiWordNet lexicon.
#         has_header (bool): Whether the lexicon has a header.
#         separator (str): The separator used in the lexicon.

#     Returns:
#         sentiwordnet (pl.DataFrame):
#           The SentiWordNet lexicon as a polars DataFrame.
#     """
#     sentiwordnet = pl.read_csv(path,
#                               has_header=has_header,
#                               schema=schema,
#                               separator=separator,
#                               # First 26 rows are comments/documentation
#                               skip_rows=27)
#     return sentiwordnet

# ---------------------------- Sentiment NRC --------------------------- #

def load_sentiment_nrc_lexicon(path: str = SENTIMENT_NRC_PATH,
                               language: str = "en",
                               has_header: bool = True,
                               separator: str = "\t",
                               **kwargs: dict[str, str],
                               ) -> pl.DataFrame:
    """
    Loads the sentiment NRC lexicon as a polars DataFrame.

    Args:
        path (str): The path to the sentiment NRC lexicon.
        schema (dict): The schema for the sentiment NRC lexicon.
        has_header (bool): Whether the lexicon has a header.
        separator (str): The separator used in the lexicon.

    Returns:
        sentiment_nrc (pl.DataFrame):
            The sentiment NRC lexicon as a polars DataFrame.
    """
    if language != "en":
        schema = SENTIMENT_NRC_SCHEMA_MULTILINGUAL
    else:
        schema = SENTIMENT_NRC_SCHEMA
    
    sentiment_nrc = pl.read_csv(path,
                               has_header=has_header,
                               schema=schema,
                               separator=separator)
    return sentiment_nrc

def filter_sentiment_lexicon(lexicon: pl.DataFrame,
                             words: list,
                             sentiment: str,
                             word_column: str = "word",
                             **kwargs: dict[str, str],
                             ) -> pl.DataFrame:
    """
    Filters the sentiment NRC lexicon for the given words and emotions.

    Args:
        lexicon (pl.DataFrame): The sentiment lexicon.
        words (list): The list of words to filter.
        sentiment (str): The sentiment to filter for.

    Returns:
        filtered_sentiment_nrc (pl.DataFrame):
            The filtered sentiment NRC lexicon.
    """
    if "label" in lexicon.columns:
        filtered_sentiment_nrc = lexicon.filter(
            (pl.col("emotion") == sentiment) &
            (pl.col("label") == 1) &
            (pl.col(word_column).is_in(words))
        )
    elif "Afrikaans" in lexicon.columns:
        filtered_sentiment_nrc = lexicon.filter(
            (pl.col(sentiment) == 1) &
            (pl.col(word_column).is_in(words))
        )
    
    return filtered_sentiment_nrc

def get_n_positive_sentiment(data: pl.DataFrame,
                            lexicon: pl.DataFrame,
                            backbone: str = "spacy",
                            nan_value: float = 0.0,
                            language: str = "en",
                            **kwargs: dict[str, str],
                            ) -> pl.DataFrame:
    """
    Calculates the number of words with positive sentiment.
    
    Args:
        data (pl.DataFrame): 
            The preprocessed input data. Contains the
            "nlp" column produced by the NLP backbone.
        lexicon (pl.DataFrame): The sentiment lexicon.
        backbone (str): The NLP backbone to use.
        nan_value (float): The value to use for NaNs.
                         Defaults to 0.0.
    
    Returns:
        data (pl.DataFrame):
            The input data with the positive sentiment count column.
            The column name is "n_positive_sentiment".
    """
    if "lemmas" not in data.columns:
        data = get_lemmas(data, backbone=backbone)

    if language != "en":
        word_column = LANGUAGES_NRC[language]
    else:
        word_column = "word"
    
    # ensure that the column is not already present;
    # this may happen if sentiment_score is called before
    # n_negative_sentiment or n_positive_sentiment
    if "n_positive_sentiment" not in data.columns:
        data = data.with_columns(
            pl.col("lemmas").map_elements(
                lambda x: filter_sentiment_lexicon(
                    lexicon, x, sentiment="positive",
                    word_column=word_column).shape[0],
                return_dtype=pl.UInt32
            ).fill_nan(nan_value).fill_null(nan_value). \
            # convention to fill NaNs with 0 as this maps to
            # the absence of positive sentiment words
                alias("n_positive_sentiment")
        )

    return data

def get_n_negative_sentiment(data: pl.DataFrame,
                            lexicon: pl.DataFrame,
                            backbone: str = "spacy",
                            nan_value: float = 0.0,
                            language: str = "en",
                            **kwargs: dict[str, str],
                            ) -> pl.DataFrame:
    """
    Calculates the number of words with negative sentiment.

    Args:
        data (pl.DataFrame):
            The preprocessed input data. Contains the
            "nlp" column produced by the NLP backbone.
        lexicon (pl.DataFrame): The sentiment lexicon.
        backbone (str): The NLP backbone to use.
        nan_value (float): The value to use for NaNs.
                            Defaults to 0.0.

    Returns:
        data (pl.DataFrame):
            The input data with the negative sentiment count column.
            The column name is "n_negative_sentiment".
    """
    if "lemmas" not in data.columns:
        data = get_lemmas(data, backbone=backbone)
    
    if language != "en":
        word_column = LANGUAGES_NRC[language]
    else:
        word_column = "word"
    
    # ensure that the column is not already present;
    # this may happen if sentiment_score is called before
    # n_negative_sentiment or n_positive_sentiment
    if "n_negative_sentiment" not in data.columns:
        data = data.with_columns(
            pl.col("lemmas").map_elements(
                lambda x: filter_sentiment_lexicon(
                    lexicon, x, sentiment="negative",
                    word_column=word_column).shape[0],
                return_dtype=pl.UInt32
            ).fill_nan(nan_value).fill_null(nan_value). \
            # convention to fill NaNs with 0 as this maps to
            # the absence of negative sentiment words
                alias("n_negative_sentiment")
        )

    return data

def get_sentiment_score(data: pl.DataFrame,
                        lexicon: pl.DataFrame,
                        backbone: str = "spacy",
                        nan_value: float = 0.0,
                        language: str = "en",
                        **kwargs: dict[str, str],
                        ) -> pl.DataFrame:
    """
    Calculates the sentiment score of the text.

    The sentiment score is calculated as the difference between the number
    of positive and negative sentiment words divided by the number of
    tokens. The sentiment score is in the range [-1, 1], where -1
    indicates negative sentiment, 0 indicates neutral sentiment, and 1 
    indicates positive sentiment.

    Args:
        data (pl.DataFrame): 
            The preprocessed input data. Contains the
            "nlp" column produced by the NLP backbone.
        lexicon (pl.DataFrame): The sentiment lexicon.
        backbone (str): The NLP backbone to use.
        nan_value (float): The value to use for NaNs.
                            Defaults to 0.0.

    Returns:
        data (pl.DataFrame): 
            The input data with the sentiment score column. The column
            name is "sentiment_score".
    """
    if language != "en":
        word_column = LANGUAGES_NRC[language]
    else:
        word_column = "word"

    if "n_positive_sentiment" not in data.columns:
        data = get_n_positive_sentiment(data,
                                        lexicon=lexicon,
                                        language=language,
                                        backbone=backbone)
    if "n_negative_sentiment" not in data.columns:
        data = get_n_negative_sentiment(data,
                                        lexicon=lexicon,
                                        language=language,
                                        backbone=backbone)
    if "n_tokens" not in data.columns:
        data = get_num_tokens(data, backbone=backbone)
    
    data = data.with_columns(
        ((pl.col("n_positive_sentiment") - pl.col("n_negative_sentiment")) /
         pl.col("n_tokens")).fill_nan(nan_value).fill_null(nan_value). \
         alias("sentiment_score")
    )

    return data




"""
This module contains functions to calculate various lexical richness
metrics from text data.

The lexical richness metrics implemented in this module are:

- Lemma/Token Ratio: 
    The ratio of the number of lemmas to the number of tokens in the text.
- Type/Token Ratio: 
    The ratio of the number of types to the number of tokens in the text.
- Root Type/Token Ratio: 
    The ratio of the number of types to the square root of the number of
    tokens in the text.
- Corrected Type/Token Ratio: 
    The ratio of the number of types to the square root of twice the 
    number of tokens in the text.
- Herdan's C: 
    The logarithm of the number of types divided by the logarithm of the
    number of tokens.
- Summer's TTR:
    The logarithm of the logarithm of the number of types divided by the
    logarithm of the logarithm of the number of tokens.
- Dougast's Uber Index:
    The square of the logarithm of the number of types divided by the
    logarithm of the number of tokens minus the logarithm of the number
    of types.
- Maas' TTR:
    The number of tokens minus the number of types divided
    by the square of the logarithm of the number of types.
- Number of Hapax Legomena:
    The number of words that occur only once in the text.
- Number of Global Token Hapax Legomena:
    The number of words that occur only once in the entire corpus.
- Number of Global Lemma Hapax Legomena:
    The number of lemmas that occur only once in the entire corpus.
- Number of Hapax Dislegomena:
    The number of words that occur only once or twice in the text.
- Number of Global Token Hapax Dislegomena:
    The number of words that occur only once or twice in the entire
    corpus.
- Number of Global Lemma Hapax Dislegomena:
    The number of lemmas that occur only once or twice in the entire
    corpus.
- Sichel's S:
    The number of hapax dislegomena divided by the number of types in the
    text.
- Global Sichel's S:
    The number of global token hapax dislegomena divided by the number of
    types in the text.
- Lexical Density:
    The ratio of the number of lexical tokens to the number of tokens in
    the text.
- Giroud's Index:
    The number of types divided by the square root of the number of tokens
    in the text.
- Measure of Textual Lexical Diversity (MTLD)
- Hypergeometric Distribution Diversity (HD-D)
- Moving-Average Type-Token Ratio (MATTR)
- Mean Segmental Type-Token Ratio (MSTTR)
- Yule's K
- Simpson's D
- Herdan's Vm
"""
from collections import Counter

import numpy as np
import polars as pl
from scipy.stats import hypergeom

from .surface import (
    get_num_tokens,
    get_num_types,
    get_token_freqs,
    get_global_token_frequencies,
    get_global_lemma_frequencies,
    get_num_lemmas
)
from .pos import (
    get_num_lexical_tokens,
)
from .preprocess import (
    get_tokens,
)

def get_lemma_token_ratio(data: pl.DataFrame,
                          backbone: str = 'spacy',
                          **kwargs: dict[str, str],
                          ) -> pl.DataFrame:
    """
    Calculates the lemma/token ratio of a text:

    N_lemmas / N_tokens.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the lemma/token ratio of the
            text data. The lemma/token ratio is stored in a new column
            named 'lemma_token_ratio'.
    """
    if 'n_tokens' not in data.columns:
        data = get_num_tokens(data, backbone=backbone)
    if 'n_lemmas' not in data.columns:
        data = get_num_lemmas(data, backbone=backbone)
    
    data = data.with_columns(
        (pl.col("n_lemmas") / pl.col("n_tokens")
         ).alias("lemma_token_ratio"),
    )

    return data

def get_ttr(data: pl.DataFrame,
            backbone: str = 'spacy',
            **kwargs: dict[str, str],
            ) -> pl.DataFrame:
    """
    Calculates the type-token ratio (TTR) of a text:

    N_types / N_tokens.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the type-token ratio of the text
            data. The type-token ratio is stored in a new column named
            'ttr'.
    """
    if 'n_tokens' not in data.columns:
        data = get_num_tokens(data, backbone=backbone)
    if 'n_types' not in data.columns:
        data = get_num_types(data, backbone=backbone)
    
    data = data.with_columns(
        (pl.col("n_types") / pl.col("n_tokens")).alias("ttr"),
    )

    return data

def get_rttr(data: pl.DataFrame,
             backbone: str = 'spacy',
             **kwargs: dict[str, str],
             ) -> pl.DataFrame:
    """
    Calculates the root type-token ratio (RTTR) of a text:

    N_types / sqrt(N_tokens).

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.
    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the root type-token ratio of the
            text data. The root type-token ratio is stored in a new column
            named 'rttr'.
    """
    if 'n_tokens' not in data.columns:
        data = get_num_tokens(data, backbone=backbone)
    if 'n_types' not in data.columns:
        data = get_num_types(data, backbone=backbone)
    
    data = data.with_columns(
        (
            pl.col("n_types") / pl.col("n_tokens") **0.5
         ).alias("rttr"),
    )

    return data

def get_cttr(data: pl.DataFrame,
             backbone: str = 'spacy',
             **kwargs: dict[str, str],
             ) -> pl.DataFrame:
    """
    Calculates the corrected type-token ratio (CTTR) of a text:

    N_types / sqrt(2*N_tokens).

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        backbone (str): The NLP library used to process the text data.
                        Either 'spacy' or 'stanza'.
    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the corrected type-token ratio
            of the text data. The corrected type-token ratio is stored in
            a new column named 'cttr'.
    """
    if 'n_tokens' not in data.columns:
        data = get_num_tokens(data, backbone=backbone)
    if 'n_types' not in data.columns:
        data = get_num_types(data, backbone=backbone)
    
    data = data.with_columns(
        (
            pl.col("n_types") / ((2 * pl.col("n_tokens")) ** 0.5)
        ).alias("cttr"),
    )

    return data

def get_herdan_c(data: pl.DataFrame,
                 backbone: str = 'spacy',
                 **kwargs: dict[str, str],
                 ) -> pl.DataFrame:
    """
    Calculates the Herdan's C of a text:

    log(N_types) / log(N_tokens).

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.
    
    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing Herdan's C of the text data.
            Herdan's C is stored in a new column named 'herdan_c'.
    """
    if 'n_tokens' not in data.columns:
        data = get_num_tokens(data, backbone=backbone)
    if 'n_types' not in data.columns:
        data = get_num_types(data, backbone=backbone)
    
    data = data.with_columns(
        (
            pl.col("n_types").log() / pl.col("n_tokens").log()
         ).fill_nan(1).alias("herdan_c")
         # convention to fill NaNs with 1 as log(1) = 0 and 
         # division by 0 is not defined.
    )

    return data

def get_summer_index(data: pl.DataFrame,
                    backbone: str = 'spacy',
                    **kwargs: dict[str, str],
                    ) -> pl.DataFrame:
    """
    Calculates the Summer's TTR of a text:

    log(log(N_types)) / log(log(N_tokens)).

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the Summer's TTR of the text
            data. Summer's TTR is stored in a new column named
            'summer_index'.
    """
    if 'n_tokens' not in data.columns:
        data = get_num_tokens(data, backbone=backbone)
    if 'n_types' not in data.columns:
        data = get_num_types(data, backbone=backbone)
    
    data = data.with_columns(
        (pl.col("n_types").log().log() / pl.col("n_tokens").log().log()
         ).fill_nan(1).alias("summer_index")
         # convention to fill NaNs with 1 as log(1) = 0 and
         # division by 0 is not defined.
    )

    return data

def get_dougast_u(data: pl.DataFrame,
                  backbone: str = 'spacy',
                  **kwargs: dict[str, str],
                  ) -> pl.DataFrame:
    """
    Calculates the Dougast's Uber index of a text:
    log(N_types)^2 / (log(N_tokens) - log(N_types)).

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.

    Returns:
        data (pl.DataFrame): A Polars DataFrame containing Dougast's Uber
        index of the text data. Dougast's Uber index is stored in a new
        column named 'dougast_u'.
    """
    if 'n_tokens' not in data.columns:
        data = get_num_tokens(data, backbone=backbone)
    if 'n_types' not in data.columns:
        data = get_num_types(data, backbone=backbone)
    
    data = data.with_columns(
        (pl.col("n_types").log()**2 / (pl.col("n_tokens").log() - \
                                       pl.col("n_types").log())
         ).fill_nan(1).alias("dougast_u")
         # convention to fill NaNs with 1 as log(1) = 0 and
         # division by 0 is not defined.
    )

    return data

def get_maas_index(data: pl.DataFrame,
                   backbone: str = 'spacy',
                   **kwargs: dict[str, str],
                   ) -> pl.DataFrame:
    """
    Calculates the Maas' TTR of a text:
    (N_tokens - N_types) / log(N_types)^2.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing Maas' TTR of the text data.
            Maas' TTR is stored in a new column named 'maas_index'.
    """
    if 'n_tokens' not in data.columns:
        data = get_num_tokens(data, backbone=backbone)
    if 'n_types' not in data.columns:
        data = get_num_types(data, backbone=backbone)
    
    data = data.with_columns(
        ((pl.col("n_tokens") - pl.col("n_types")) / \
            pl.col("n_types").log()**2 
         ).fill_nan(1).alias("maas_index")
         # convention to fill NaNs with 1 as log(1) = 0 and
         # division by 0 is not defined.
    )

    return data

def get_n_hapax_legomena(data: pl.DataFrame,
                         backbone: str = 'spacy',
                         **kwargs: dict[str, str],
                         ) -> pl.DataFrame:
    """
    Calculates the number of hapax legomena in a text: words that occur
    only once.

    TODO: Add global hapax legomena calculation.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the number of hapax legomena in
            the text data. The number of hapax legomena is stored in a new
            column named 'n_hapax_legomena'.
    """
    if backbone == 'spacy':
        data = data.with_columns(
             pl.col("nlp").map_elements(lambda x: np.sum(
                  np.unique(np.array([token.text for token in x]),
                            return_counts=True)[1] == 1),
                            return_dtype=pl.UInt32).alias("n_hapax_legomena")
        )

    elif backbone == 'stanza':
        data = data.with_columns(
            pl.col("nlp").map_elements(lambda x: np.sum(
                np.unique(np.array([token.text 
                                    for sent 
                                    in x.sentences 
                                    for token in sent.tokens]),
                          return_counts=True)[1] == 1),
                          return_dtype=pl.UInt32).alias("n_hapax_legomena")
        )
    
    return data

def get_n_global_token_hapax_legomena(data: pl.DataFrame,
                                      backbone: str = 'spacy',
                                      **kwargs: dict[str, str],
                                      ) -> pl.DataFrame:
    """
    Calculates the number of global hapax legomena in a text: words that
    occur only once in the entire corpus.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.
    
    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the number of global hapax
            legomena in the text data. The number of global hapax legomena
            is stored in a new column named 'n_global_hapax_legomena'.
    """
    token_freqs = get_global_token_frequencies(data, backbone=backbone)
    if backbone == 'spacy':
        data = data.with_columns(
            pl.col("nlp").map_elements(lambda x: len(
                [token for token in x if token_freqs[token.text] == 1]),
                return_dtype=pl.UInt32). \
                    alias("n_global_token_hapax_legomena")
        )
    elif backbone == 'stanza':
        data = data.with_columns(
            pl.col("nlp").map_elements(lambda x: len(
                [token for sent in x.sentences for token in sent.tokens
                 if token_freqs[token.text] == 1]),
                return_dtype=pl.UInt32). \
                    alias("n_global_token_hapax_legomena")
        )

    return data

def get_n_global_lemma_hapax_legomena(data: pl.DataFrame,
                                      backbone: str = 'spacy',
                                      **kwargs: dict[str, str],
                                      ) -> pl.DataFrame:
    """
    Calculates the number of global hapax legomena in a text: lemmas that
    occur only once in the entire corpus.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the number of global hapax
            legomena in the text data. The number of global hapax legomena
            is stored in a new column named 'n_global_lemma_hapax_legomena'.
    """
    lemma_freqs = get_global_lemma_frequencies(data, backbone=backbone)
    if backbone == 'spacy':
        data = data.with_columns(
            pl.col("nlp").map_elements(lambda x: len(
                [token.lemma_ for token in x if 
                 lemma_freqs[token.lemma_] == 1]),
                return_dtype=pl.UInt32). \
                    alias("n_global_lemma_hapax_legomena")
        )
    elif backbone == 'stanza':
        data = data.with_columns(
            pl.col("nlp").map_elements(lambda x: len(
                [token.lemma for sent in x.sentences for 
                 token in sent.tokens
                 if lemma_freqs[token.lemma] == 1]),
                return_dtype=pl.UInt32). \
                    alias("n_global_lemma_hapax_legomena")
        )

    return data

def get_n_hapax_dislegomena(data: pl.DataFrame,
                            backbone: str = 'spacy',
                            **kwargs: dict[str, str],
                            ) -> pl.DataFrame:
    """
    Calculates the number of hapax dislegomena in a text: words that occur
    only once or twice.

    TODO: Add global hapax dislegomena calculation.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the number of hapax dislegomena
            in the text data. The number of hapax dislegomena is stored in
            a new column named 'n_hapax_dislegomena'.
    """
    if backbone == 'spacy':
        data = data.with_columns(
             pl.col("nlp").map_elements(lambda x: np.sum(
                  np.unique(np.array([token.text for token in x]),
                            return_counts=True)[1] <= 2),
                            return_dtype=pl.UInt32). \
                                alias("n_hapax_dislegomena")
        )

    elif backbone == 'stanza':
        data = data.with_columns(
            pl.col("nlp").map_elements(lambda x: np.sum(
                np.unique(np.array([token.text 
                                    for sent 
                                    in x.sentences 
                                    for token in sent.tokens]),
                          return_counts=True)[1] <= 2),
                          return_dtype=pl.UInt32). \
                            alias("n_hapax_dislegomena")
        )
    
    return data

def get_n_global_token_hapax_dislegomena(data: pl.DataFrame,
                                       backbone: str = 'spacy',
                                       **kwargs: dict[str, str],
                                       ) -> pl.DataFrame:
    """
    Calculates the number of global hapax dislegomena in a text: words that
    occur only once or twice in the entire corpus.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        backbone (str): The NLP library used to process the text data.
                        Either 'spacy' or 'stanza'.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the number of global hapax
            dislegomena in the text data. The number of global hapax
            dislegomena is stored in a new column named
            'n_global_token_hapax_dislegomena'.
    """
    token_freqs = get_global_token_frequencies(data, backbone=backbone)

    if backbone == 'spacy':
        data = data.with_columns(
            pl.col("nlp").map_elements(lambda x: len(
                [token for token in x if token_freqs[token.text] <= 2]),
                return_dtype=pl.UInt32). \
                    alias("n_global_token_hapax_dislegomena")
        )
    elif backbone == 'stanza':
        data = data.with_columns(
            pl.col("nlp").map_elements(lambda x: len(
                [token for sent in x.sentences for token in sent.tokens
                 if token_freqs[token.text] <= 2]),
                return_dtype=pl.UInt32). \
                    alias("n_global_token_hapax_dislegomena")
        )

    return data

def get_n_global_lemma_hapax_dislegomena(data: pl.DataFrame,
                                       backbone: str = 'spacy',
                                       **kwargs: dict[str, str],
                                       ) -> pl.DataFrame:
    """
    Calculates the number of global hapax dislegomena in a text: lemmas that
    occur only once or twice in the entire corpus.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        backbone (str): The NLP library used to process the text data.
                        Either 'spacy' or 'stanza'.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the number of global hapax
            dislegomena in the text data. The number of global hapax
            dislegomena is stored in a new column named
            'n_global_lemma_hapax_dislegomena'.
    """
    lemma_freqs = get_global_lemma_frequencies(data, backbone=backbone)

    if backbone == 'spacy':
        data = data.with_columns(
            pl.col("nlp").map_elements(lambda x: len(
                [token.lemma_ for token in x if 
                 lemma_freqs[token.lemma_] <= 2]),
                return_dtype=pl.UInt32). \
                    alias("n_global_lemma_hapax_dislegomena")
        )
    elif backbone == 'stanza':
        data = data.with_columns(
            pl.col("nlp").map_elements(lambda x: len(
                [token.lemma for sent in x.sentences for 
                 token in sent.tokens
                 if lemma_freqs[token.lemma] <= 2]),
                return_dtype=pl.UInt32). \
                    alias("n_global_lemma_hapax_dislegomena")
        )

    return data

def get_sichel_s(data: pl.DataFrame,
                 backbone: str = 'spacy',
                 **kwargs: dict[str, str],
                 ) -> pl.DataFrame:
    """
    Calculates the Sichel's S of a text:
    N_hapax_dislegomena / N_types.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.
    
    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing Sichel's S of the text data.
            Sichel's S is stored in a new column named 'sichel_s'.
    """
    if 'n_types' not in data.columns:
        data = get_num_types(data, backbone=backbone)
    if 'n_hapax_dislegomena' not in data.columns:
        data = get_n_hapax_dislegomena(data, backbone=backbone)

    data = data.with_columns(
        (pl.col("n_hapax_dislegomena") / pl.col("n_types")
         ).alias("sichel_s"),
    )

    return data

def get_global_sichel_s(data: pl.DataFrame,
                        backbone: str = 'spacy',
                        **kwargs: dict[str, str],
                        ) -> pl.DataFrame:
    """
    Calculates the global Sichel's S of a text:
    N_global_token_hapax_dislegomena / N_types.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        backbone (str): The NLP library used to process the text data.
                        Either 'spacy' or 'stanza'.
    
    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the global Sichel's S of the
            text data. The global Sichel's S is stored in a new column
            named 'global_sichel_s'.
    """
    if 'n_types' not in data.columns:
        data = get_num_types(data, backbone=backbone)
    if 'n_global_token_hapax_dislegomena' not in data.columns:
        data = get_n_global_token_hapax_dislegomena(data,
                                                    backbone=backbone)

    data = data.with_columns(
        (pl.col("n_global_token_hapax_dislegomena") / pl.col("n_types")
         ).alias("global_sichel_s"),
    )

    return data

def get_lexical_density(data: pl.DataFrame,
                        backbone: str = 'spacy',
                        **kwargs: dict[str, str],
                        ) -> pl.DataFrame:
    """
    Calculates the lexical density of a text:

    N_lex / N_tokens.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the lexical density of the text
            data. The lexical density is stored in a new column named
            'lexical_density'.
    """
    if 'n_tokens' not in data.columns:
        data = get_num_tokens(data, backbone=backbone)
    if 'n_lexical_tokens' not in data.columns:
        data = get_num_lexical_tokens(data, backbone=backbone)
    
    data = data.with_columns(
        (pl.col("n_lexical_tokens") / pl.col("n_tokens")
         ).alias("lexical_density"),
    )

    return data

def get_giroud_index(data: pl.DataFrame,
                     backbone: str = 'spacy',
                     **kwargs: dict[str, str],
                     ) -> pl.DataFrame:
    """
    Calculates the Giroud's index of a text:

    N_types / sqrt(N_tokens).

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing Giroud's C of the text data.
            Giroud's C is stored in a new column named 'giroud_c'.
    """
    if 'n_tokens' not in data.columns:
        data = get_num_tokens(data, backbone=backbone)
    if 'n_types' not in data.columns:
        data = get_num_types(data, backbone=backbone)
    
    data = data.with_columns(
        (pl.col("n_types") / pl.col("n_tokens").sqrt()
         ).alias("giroud_index"),
    )

    return data

def get_mtld(data: pl.DataFrame,
             threshold: float = 0.72,
             backbone: str = 'spacy',
             **kwargs: dict[str, str],
             ) -> pl.DataFrame:
    """
    Calculates the Measure of Textual Lexical Diversity (MTLD) of a text.

    For reference https://link.springer.com/article/10.3758/BRM.42.2.381 

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.
        threshold: The threshold value for the MTLD.
                 The default value is 0.72.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the MTLD of the text data.
            The MTLD is stored in a new column named 'mtld'.
    """
    def sub_mtld(tokens: list[str],
                 forward: bool = True,
                 threshold: float = 0.72
                 ) -> float:
        """
        Calculate the MTLD of a text using a forward or backward approach.
        For reference https://link.springer.com/article/10.3758/BRM.42.2.381 
        """
        if not forward:
            tokens = tokens[::-1]
        
        unique_tokens = set()
        current_ttr = 1.0
        factors = 0.0
        token_count = 0

        for i, token in enumerate(tokens):
            token_count += 1
            unique_tokens.add(token)
            current_ttr = len(unique_tokens) / token_count

            if i == len(tokens) - 1 and current_ttr >= threshold:
                factors += (current_ttr - 1) / (threshold - 1)
            elif current_ttr < threshold:
                factors += 1
                # Reset
                unique_tokens = set()
                current_ttr = 1.0
                token_count = 0

        return len(tokens) / factors if factors != 0 else len(tokens)
            
    if 'tokens' not in data.columns:
        data = get_tokens(data, backbone=backbone)
    
    data = data.with_columns(
        pl.col("tokens").map_elements(lambda x: (
            sub_mtld(x, threshold=threshold) + \
                sub_mtld(x, forward=False, threshold=threshold)
            ) / 2.0, return_dtype=pl.Float32).alias("mtld")
    )

    return data


def get_hdd(data: pl.DataFrame,
            backbone: str = 'spacy',
            draws: int = 42,
            **kwargs: dict[str, str],
            ) -> pl.DataFrame:
    """
    Calculates the Hypergeometric Distribution Diversity (HD-D) of a text.

    The default number of draws is 42. We note, however, that this value
    should be smaller than the number of tokens in the text and thus will
    need to be adjusted for most short texts. A number of draws that is
    too large will result in NaN values.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.
        draws: The number of draws for the HDD. The default value is 42.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the HD-D of the text data. 
            The HD-D is stored in a new column named 'hdd'.
    """
    def hdd(tokens: list[str],
            draws: int = 42,
            ) -> float:
        """
        Calculate the HDD of a text.
        """
        freqs = Counter(tokens)
        n_types = len(freqs)
        n_tokens = len(tokens)
        hdd = 0.0

        for freq in freqs.values():
            hdd += hypergeom.pmf(freq, n_tokens, n_types, draws)
        
        return hdd
    
    if 'tokens' not in data.columns:
        data = get_tokens(data, backbone=backbone)
    
    data = data.with_columns(
        pl.col("tokens").map_elements(lambda x: hdd(x, draws=draws),
                                      return_dtype=pl.Float32).alias("hdd")
    )

    return data

def get_mattr(data: pl.DataFrame,
              backbone: str = 'spacy',
              window_size: int = 5,
              **kwargs: dict[str, str],
              ) -> pl.DataFrame:
    """
    Calculates the Moving-Average Type-Token Ratio (MATTR) of a text.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.
        window_size (int):
            The size of the window for the MATTR calculation.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the MATTR of the text data.
            The MATTR is stored in a new column named 'mattr'.
    """
    def mattr(tokens: list[str],
              window_size: int = 5,
              ) -> float:
        """
        Calculate the MATTR of a text.
        """
        n_types = []
        n_tokens = []
        for i in range(0, len(tokens)):
            window = tokens[i:i+window_size]
            n_types.append(len(set(window)))
            n_tokens.append(len(window))
        
        return np.mean(n_types) / np.mean(n_tokens)
    
    if 'tokens' not in data.columns:
        data = get_tokens(data, backbone=backbone)
    
    data = data.with_columns(
        pl.col("tokens").map_elements(
            lambda x: mattr(x,
                            window_size=window_size),
                            return_dtype=pl.Float32).alias("mattr")
    )

    return data

def get_msttr(data: pl.DataFrame,
              backbone: str = 'spacy',
              window_size: int = 5,
              discard: bool = True,
              **kwargs: dict[str, str],
              ) -> pl.DataFrame:
    """
    Calculates the Mean Segmental Type-Token Ratio (MSTTR) of a text.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        backbone (str):
            The NLP library used to process the text data. Either 'spacy'
            or 'stanza'.
        window_size (int):
            The size of the window for the MSTTR calculation.
        discard (bool):
            Whether to discard the last window if it is not complete.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the MSTTR of the text data.
            The MSTTR is stored in a new column named 'msttr'.
    """
    def msttr(tokens: list[str],
              window_size: int = 5,
              discard: bool = True,
              ) -> float:
        """
        Calculate the MSTTR of a text.
        """
        scores = []

        for i in range(0, len(tokens), window_size):
            window = tokens[i:i+window_size]
            n_types = len(set(window))
            n_tokens = len(window)
            scores.append(n_types / n_tokens)
        
        # Discard the last window if it is not complete
        if discard and len(tokens) % window_size != 0:
            scores = scores[:-1]

        return np.mean(scores)
    if 'tokens' not in data.columns:
        data = get_tokens(data, backbone=backbone)
    
    data = data.with_columns(
        pl.col("tokens"). \
            map_elements(lambda x: msttr(x,
                                         window_size=window_size,
                                         discard=discard),
                                         return_dtype=pl.Float32). \
                                            alias("msttr")
    )

    return data

def get_yule_k(data: pl.DataFrame,
               backbone: str = 'spacy',
               **kwargs: dict[str, str],
               ) -> pl.DataFrame:
    """
    Calculates the Yule's K of a text.
    Yule's K is a characteristic of the vocabulary richness of a text.
    It is calculated as:

    K = 10^4 * (Σ(V(i,N) * (i/n)^2) - n) / n^2
    = 10^4 * ((Σ(V(i,N) * (i^2/n^4)) - (1/n))

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        backbone (str): The NLP library used to process the text data.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing Yule's K of the text data.
            Yule's K is stored in a new column named 'yule_k'.
    """
    def inner_sum(x: dict[str, int]) -> float:
        """
        Calculate the inner sum of the Yule's K formula, i.e.:
        (Σ(V(i,N) * (i^2/n^4))
        """
        counts = Counter(x.values())
        # gathering n from counts to streamline the calculation
        n = sum(counts.values())
        n_power_4 = n ** 4  # Precompute to avoid redundant calculations

        inner = sum(count * (i ** 2 / n_power_4) for i, count in \
                    counts.items())

        return inner
    
    # get the number of tokens
    if 'n_tokens' not in data.columns:
        data = get_num_tokens(data, backbone=backbone)
    # calculate frequency of each token per text
    if 'token_freqs' not in data.columns:
        data = get_token_freqs(data, backbone=backbone)

    data = data.with_columns(
        (pl.lit(10**4) * \
        pl.col("token_freqs").map_elements(lambda x: inner_sum(x),
                                           return_dtype=pl.Float32) - \
        (1 / pl.col("n_tokens"))).alias("yule_k")
    )

    return data


def get_simpsons_d(data: pl.DataFrame,
                   backbone: str = 'spacy',
                   **kwargs: dict[str, str],
                   ) -> pl.DataFrame:
    """
    Calculates the Simpson's D of a text.
    
    D = Σ(V(i,N) * (i/n) * ((i-1)/(n-1)))

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing Simpson's D of the text data.
            Simpson's D is stored in a new column named 'simpsons_d'.
    """
    def simpsons_d(x: dict[str, int]) -> float:
        """
        Calculate the Simpson's D of a text.
        """
        counts = Counter(x.values())
        n = sum(counts.values())
        n_inv = 1 / n  # Precompute 1 / n
        n_minus_1_inv = 1 / (n - 1)  # Precompute 1 / (n - 1)

        d = sum(count * i * n_inv * (i - 1) * n_minus_1_inv for \
                i, count in counts.items())

        return d
    if 'token_freqs' not in data.columns:
        data = get_token_freqs(data, backbone=backbone)
    
    data = data.with_columns(
        pl.col("token_freqs").map_elements(lambda x: simpsons_d(x),
                                           return_dtype=pl.Float32). \
                                            alias("simpsons_d")
    )

    return data
    

def get_herdan_v(data: pl.DataFrame,
                 backbone: str = 'spacy',
                 **kwargs: dict[str, str],
                 ) -> pl.DataFrame:
    """
    Calculates the Herdan's Vm of a text:

    Vm^2 = K + (1/N) - (1/V(N))
    Vm = sqrt(K + (1/N) - (1/V(N)))

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.
    
    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing Herdan's V of the text data. 
            Herdan's V is stored in a new column named 'herdan_v'.
    """
    if not 'n_tokens' in data.columns:
        data = get_num_tokens(data, backbone=backbone)
    if not 'n_types' in data.columns:
        data = get_num_types(data, backbone=backbone)
    if not 'yule_k' in data.columns:
        data = get_yule_k(data, backbone=backbone)

    data = data.with_columns(
        (pl.col("yule_k") + (1 / pl.col("n_tokens")) - \
         (1 / pl.col("n_types")).sqrt()
        ).alias("herdan_v")
    )

    return data




"""
This module contains schema definitions for the different external
resources. These schemas are used in Polars to enable type-safe and
fast data manipulation.
"""
import polars as pl

# ======================================== #
#              General Blocks              #
# ======================================== #

LANGS_NRC = {
    'Afrikaans': pl.String,
    'Albanian': pl.String,
    'Amharic': pl.String,
    'Arabic': pl.String,
    'Armenian': pl.String,
    'Azerbaijani': pl.String,
    'Basque': pl.String,
    'Belarusian': pl.String,
    'Bengali': pl.String,
    'Bosnian': pl.String,
    'Bulgarian': pl.String,
    'Catalan': pl.String,
    'Cebuano': pl.String,
    'Chichewa': pl.String,
    'Chinese-Simplified': pl.String,
    'Chinese-Traditional': pl.String,
    'Corsican': pl.String,
    'Croatian': pl.String,
    'Czech': pl.String,
    'Danish': pl.String,
    'Dutch': pl.String,
    'Esperanto': pl.String,
    'Estonian': pl.String,
    'Filipino': pl.String,
    'Finnish': pl.String,
    'French': pl.String,
    'Frisian': pl.String,
    'Galician': pl.String,
    'Georgian': pl.String,
    'German': pl.String,
    'Greek': pl.String,
    'Gujarati': pl.String,
    'Haitian-Creole': pl.String,
    'Hausa': pl.String,
    'Hawaiian': pl.String,
    'Hebrew': pl.String,
    'Hindi': pl.String,
    'Hmong': pl.String,
    'Hungarian': pl.String,
    'Icelandic': pl.String,
    'Igbo': pl.String,
    'Indonesian': pl.String,
    'Irish': pl.String,
    'Italian': pl.String,
    'Japanese': pl.String,
    'Javanese': pl.String,
    'Kannada': pl.String,
    'Kazakh': pl.String,
    'Khmer': pl.String,
    'Kinyarwanda': pl.String,
    'Korean': pl.String,
    'Kurdish-Kurmanji': pl.String,
    'Kyrgyz': pl.String,
    'Lao': pl.String,
    'Latin': pl.String,
    'Latvian': pl.String,
    'Lithuanian': pl.String,
    'Luxembourgish': pl.String,
    'Macedonian': pl.String,
    'Malagasy': pl.String,
    'Malay': pl.String,
    'Malayalam': pl.String,
    'Maltese': pl.String,
    'Maori': pl.String,
    'Marathi': pl.String,
    'Mongolian': pl.String,
    'Myanmar-Burmese': pl.String,
    'Nepali': pl.String,
    'Norwegian': pl.String,
    'Odia': pl.String,
    'Pashto': pl.String,
    'Persian': pl.String,
    'Polish': pl.String,
    'Portuguese': pl.String,
    'Punjabi': pl.String,
    'Romanian': pl.String,
    'Russian': pl.String,
    'Samoan': pl.String,
    'Scots-Gaelic': pl.String,
    'Serbian': pl.String,
    'Sesotho': pl.String,
    'Shona': pl.String,
    'Sindhi': pl.String,
    'Sinhala': pl.String,
    'Slovak': pl.String,
    'Slovenian': pl.String,
    'Somali': pl.String,
    'Spanish': pl.String,
    'Sundanese': pl.String,
    'Swahili': pl.String,
    'Swedish': pl.String,
    'Tajik': pl.String,
    'Tamil': pl.String,
    'Tatar': pl.String,
    'Telugu': pl.String,
    'Thai': pl.String,
    'Turkish': pl.String,
    'Turkmen': pl.String,
    'Ukrainian': pl.String,
    'Urdu': pl.String,
    'Uyghur': pl.String,
    'Uzbek': pl.String,
    'Vietnamese': pl.String,
    'Welsh': pl.String,
    'Xhosa': pl.String,
    'Yiddish': pl.String,
    'Yoruba': pl.String,
    'Zulu': pl.String,
}

# ======================================== #
#        Sentiment/Emotion Analysis        #
# ======================================== #

# ---------------------------------------- #
#                   VAD                    #
# ---------------------------------------- #

VAD_SCHEMA_NRC = {
    "word": pl.String,
    "valence": pl.Float32,
    "arousal": pl.Float32,
    "dominance": pl.Float32,
}

VAD_SCHEMA_NRC_MULTILINGUAL = {
    'English Word': pl.String,
    'Valence': pl.Float32,
    'Arousal': pl.Float32,
    'Dominance': pl.Float32,
}.update(LANGS_NRC)

VAD_SCHEMA_WARRINER = {
    '': pl.UInt32,
    'Word': pl.String,
    "V.Mean.Sum": pl.Float32,
    "V.SD.Sum": pl.Float32,
    "V.Rat.Sum": pl.Float32,
    "A.Mean.Sum": pl.Float32,
    "A.SD.Sum": pl.Float32,
    "A.Rat.Sum": pl.Float32,
    "D.Mean.Sum": pl.Float32,
    "D.SD.Sum": pl.Float32,
    "D.Rat.Sum": pl.Float32,
    "V.Mean.M": pl.Float32,
    "V.SD.M": pl.Float32,
    "V.Rat.M": pl.Float32,
    "V.Mean.F": pl.Float32,
    "V.SD.F": pl.Float32,
    "V.Rat.F": pl.Float32,
    "A.Mean.M": pl.Float32,
    "A.SD.M": pl.Float32,
    "A.Rat.M": pl.Float32,
    "A.Mean.F": pl.Float32,
    "A.SD.F": pl.Float32,
    "A.Rat.F": pl.Float32,
    "D.Mean.M": pl.Float32,
    "D.SD.M": pl.Float32,
    "D.Rat.M": pl.Float32,
    "D.Mean.F": pl.Float32,
    "D.SD.F": pl.Float32,
    "D.Rat.F": pl.Float32,
    "V.Mean.Y": pl.Float32,
    "V.SD.Y": pl.Float32,
    "V.Rat.Y": pl.Float32,
    "V.Mean.O": pl.Float32,
    "V.SD.O": pl.Float32,
    "V.Rat.O": pl.Float32,
    "A.Mean.Y": pl.Float32,
    "A.SD.Y": pl.Float32,
    "A.Rat.Y": pl.Float32,
    "A.Mean.O": pl.Float32,
    "A.SD.O": pl.Float32,
    "A.Rat.O": pl.Float32,
    "D.Mean.Y": pl.Float32,
    "D.SD.Y": pl.Float32,
    "D.Rat.Y": pl.Float32,
    "D.Mean.O": pl.Float32,
    "D.SD.O": pl.Float32,
    "D.Rat.O": pl.Float32,
    "V.Mean.L": pl.Float32,
    "V.SD.L": pl.Float32,
    "V.Rat.L": pl.Float32,
    "V.Mean.H": pl.Float32,
    "V.SD.H": pl.Float32,
    "V.Rat.H": pl.Float32,
    "A.Mean.L": pl.Float32,
    "A.SD.L": pl.Float32,
    "A.Rat.L": pl.Float32,
    "A.Mean.H": pl.Float32,
    "A.SD.H": pl.Float32,
    "A.Rat.H": pl.Float32,
    "D.Mean.L": pl.Float32,
    "D.SD.L": pl.Float32,
    "D.Rat.L": pl.Float32,
    "D.Mean.H": pl.Float32,
    "D.SD.H": pl.Float32,
    "D.Rat.H": pl.Float32,
}

# ---------------------------------------- #
#           Emotion Intensity              #
# ---------------------------------------- #

INTENSITY_SCHEMA = {
    "word": pl.String,
    "emotion": pl.String,
    "emotion_intensity": pl.Float32,
}

INTENSITY_SCHEMA_MULTILINGUAL = {
    'English Word': pl.String,
    'Emotion': pl.String,
    'Emotion-Intensity Score': pl.Float32,
}.update(LANGS_NRC)

# ---------------------------------------- #
#               Sentiment                  #
# ---------------------------------------- #

SENTIWORDNET_SCHEMA = {
    "POS": pl.String,
    "ID": pl.UInt32,
    "PosScore": pl.Float32,
    "NegScore": pl.Float32,
    "SynsetTerms": pl.String,
    "Gloss": pl.String,
}

SENTIMENT_NRC_SCHEMA = {
    "word": pl.String,
    "emotion": pl.String,
    "label": pl.UInt8,
}

SENTIMENT_NRC_SCHEMA_MULTILINGUAL = {
    "English Word": pl.String,
    "anger": pl.UInt8,
    "anticipation": pl.UInt8,
    "disgust": pl.UInt8,
    "fear": pl.UInt8,
    "joy": pl.UInt8,
    "negative": pl.UInt8,
    "positive": pl.UInt8,
    "sadness": pl.UInt8,
    "surprise": pl.UInt8,
    "trust": pl.UInt8,
}.update(LANGS_NRC)

# ======================================== #
#        Psycholinguistic Features         #
# ======================================== #


CONFIG_ALL = {
    "backbone": "spacy",
    "language": "en",
    "model": "en_core_web_sm",
    "text_column": "text",
    "max_length": 1000000,
    "remove_constant_cols": True,
    "features": {
        "surface": [
            "raw_sequence_length",
            "n_tokens",
            "n_sentences",
            "n_tokens_per_sentence",
            "n_characters",
            "avg_word_length",
            "n_types",
            "n_long_words",
            "n_lemmas"
        ],
        "morphology": ["n_per_morph_feature"],
        "dependency": [
            "tree_width",
            "tree_depth",
            "tree_branching",
            "n_noun_chunks",
            "n_per_dependency_type"
        ],
        "emotion": [
            "sentiment_score",
            "n_negative_sentiment",
            "n_positive_sentiment",
            "avg_valence",
            "n_low_valence",
            "n_high_valence",
            "avg_arousal",
            "n_low_arousal",
            "n_high_arousal",
            "avg_dominance",
            "n_low_dominance",
            "n_high_dominance",
            "avg_emotion_intensity",
            "n_low_intensity",
            "n_high_intensity",
        ],
        "information": [
            "compressibility",
            "entropy"
        ],
        "lexical_richness": [
            "lemma_token_ratio",
            "ttr",
            "rttr",
            "cttr",
            "herdan_c",
            "summer_index",
            "dougast_u",
            "maas_index",
            "lexical_density",
            "n_hapax_legomena",
            "n_global_token_hapax_legomena",
            "n_global_lemma_hapax_legomena",
            "n_hapax_dislegomena",
            "n_global_token_hapax_dislegomena",
            "n_global_lemma_hapax_dislegomena",
            "sichel_s",
            "global_sichel_s"
            "giroud_index",
            "mtld",
            "hdd",
            "mattr",
            "msttr",
            "yule_k",
            "simpsons_d",
            "herdan_v",
        ],
        "pos": [
            "n_lexical_tokens",
            "pos_variability",
            "n_per_pos",
        ],
        "psycholinguistic": [
            "avg_concreteness",
            "n_low_concreteness",
            "n_high_concreteness",
            "avg_sd_concreteness",
            "n_controversial_concreteness",
            "avg_aoa",
            "n_low_aoa",
            "n_high_aoa",
            "avg_sd_aoa",
            "n_controversial_aoa",
            "avg_prevalence",
            "n_low_prevalence",
            "n_high_prevalence",
            "avg_socialness",
            "n_low_socialness",
            "n_high_socialness",
            "avg_sd_socialness",
            "n_controversial_socialness",
            "avg_sensorimotor",
            "n_low_sensorimotor",
            "n_high_sensorimotor",
            "avg_sd_sensorimotor",
            "n_controversial_sensorimotor",
            "avg_iconicity",
            "n_low_iconicity",
            "n_high_iconicity",
            "avg_sd_iconicity",
            "n_controversial_iconicity",
        ],
        "readability": [
            "n_syllables",
            "n_monosyllables",
            "n_polysyllables",
            "flesch_reading_ease",
            "flesch_kincaid_grade",
            "gunning_fog",
            "ari",
            "smog",
            "cli",
            "lix",
            "rix",
        ],
        "semantic": [
            "n_lemmas",
            "n_hedges",
            "avg_num_synsets",
            "avg_num_synsets_per_pos",
            "n_low_synsets",
            "n_high_synsets",
            "n_high_synsets_per_pos",
            "n_low_synsets_per_pos",
        ],
        "entities": [
            "n_entities",
            "n_entities_token_ratio",
            "n_entities_sentence_ratio",
            "n_per_entity_type",
        ],
    }
}


"""
This module contains functions for calculating information-theoretic
metrics.

The information-theoretic metrics implemented in this module are:

- Compressibility:
    The ratio of the length of the compressed text to the length of the
    original text.
- Shannon Entropy:
    The Shannon entropy of the text data.
"""
import bz2

import numpy as np
import polars as pl

def get_compressibility(data: pl.DataFrame,
                        text_column: str = 'text',
                        **kwargs: dict[str, str],
                        ) -> pl.DataFrame:
    """
    Calculates the compressibility of the texts in the text column.

    The compressibility is the ratio of the length of the compressed text
    to the length of the original text. This is used as a proxy for the
    Kolmogorov complexity of the text.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        text_column (str): The name of the column containing the text data.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the compressibility of the text
            data. The compressibility is stored in a new column named
            'compressibility'.
    """
    data = data.with_columns(
        pl.col(text_column).map_elements(
            lambda x: len(bz2.compress(x.encode('utf-8'))) / len(x),
            return_dtype=pl.Float64
            ).alias("compressibility"),
    )
    
    return data

def get_entropy(data: pl.DataFrame,
                text_column: str = 'text',
                **kwargs: dict[str, str],
                ) -> pl.DataFrame:
    """
    Calculates the Shannon entropy of the texts in the text column.

    The Shannon entropy is a measure of the uncertainty in a random variable.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        text_column (str): The name of the column containing the text data.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the Shannon entropy of the text
            data. The Shannon entropy is stored in a new column named
            'entropy'.
    """
    data = data.with_columns(
        pl.col(text_column).map_elements(
            entropy,
            return_dtype=pl.Float64
            ).alias("entropy"),
    )

    return data

def entropy(string: str,
            **kwargs: dict[str, str],
            ) -> float:
    """
    Calculate the Shannon entropy of a string.
    Helper function for get_entropy.

    Args:
        string( str): The input string.

    Returns:
        entropy (float): The Shannon entropy of the input string.
    """
    chars = np.array(list(string))

    _, counts = np.unique(chars, return_counts=True)
    prob = counts / len(string)
    entropy = -np.sum(prob * np.log2(prob))
    
    return entropy




"""
This module contains functions to calculate various part-of-speech (POS)-
related features from text data.

The POS-related features implemented in this module are:

- Number of lexical tokens: The number of tokens that are nouns, verbs,
                            adjectives, or adverbs.
- POS variability: The ratio of the number of unique POS tags to the
                   number of tokens.
- Number of tokens per POS tag: The number of tokens for each POS tag.
"""
import polars as pl

from .surface import (
    get_num_tokens,
)

UPOS_TAGS = [
    'ADJ', 'ADP', 'ADV', 'AUX', 'CONJ', 'CCONJ', 'DET', 'INTJ', 'NOUN',
    'NUM', 'PART', 'PRON', 'PROPN', 'PUNCT', 'SCONJ', 'SYM', 'VERB', 'X'
]

def get_num_lexical_tokens(data: pl.DataFrame,
                           backbone: str = 'spacy',
                           **kwargs: dict[str, str],
                           ) -> pl.DataFrame:
    """
    Calculates the number of lexical tokens in the text.

    Lexical tokens are tokens that are nouns, verbs, adjectives, or adverbs.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        backbone (str): The NLP library used to process the text data.
                        Either 'spacy' or 'stanza'.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the number of lexical tokens
            in the text data. The number of lexical tokens is stored in
            a new column named 'n_lexical_tokens'.
    """
    lex = ["NOUN", "VERB", "ADJ", "ADV"]
    if backbone == 'spacy':
        data = data.with_columns(
            pl.col("nlp").map_elements(lambda x: len(
                [token for token in x if token.pos_ in lex]),
                return_dtype=pl.UInt16
                ).alias("n_lexical_tokens"),
        )
    elif backbone == 'stanza':
        data = data.with_columns(
            pl.col("nlp").map_elements(lambda x: len(
                [token for sent in x.sentences for token
                 in sent.words if token.upos in lex]),
                return_dtype=pl.UInt16
                ).alias("n_lexical_tokens"),
        )
    
    return data

def get_pos_variability(data: pl.DataFrame,
                        backbone: str = 'spacy',
                        **kwargs: dict[str, str],
                        ) -> pl.DataFrame:
    """
    Calculates the variability of part-of-speech tags in the text data.

    The variability is the ratio of the number of unique part-of-speech
    tags to the number of tokens in the text data.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        backbone (str): The NLP library used to process the text data.
                        Either 'spacy' or 'stanza'.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the variability of
            part-of-speech tags in the text data. The variability
            is stored in a new column named  'pos_variability'.
    """
    if "n_tokens" not in data.columns:
        data = get_num_tokens(data, backbone)
    
    if backbone == 'spacy':
        data = data.with_columns(
            (pl.col("nlp").map_elements(lambda x: len(set(
                [token.pos_ for token in x])),
                return_dtype=pl.UInt16) / 
            pl.col("n_tokens")).alias("pos_variability"),
        )
    elif backbone == 'stanza':
        data = data.with_columns(
            (pl.col("nlp").map_elements(lambda x: len(set(
                [token.upos for sent in x.sentences for token
                 in sent.words])),
                return_dtype=pl.UInt16) / 
            pl.col("n_tokens")).alias("pos_variability"),
        )

    return data

def get_num_per_pos(data: pl.DataFrame,
                    backbone: str = 'spacy',
                    pos_tags: list = UPOS_TAGS,
                    **kwargs: dict[str, str],
                    ) -> pl.DataFrame:
    """
    Calculates the number of tokens per part-of-speech tag in the text data.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.
        pos_tags: A list of part-of-speech tags to calculate the number of
                tokens for.
                Default is the Universal POS tagset.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the number of tokens per
            part-of-speech tag in the text data. The number of tokens
            per part-of-speech tag is stored in new columns named
            'n_{pos}' where {pos} is the part-of speech tag.
    """
    if backbone == 'spacy':
        for pos in pos_tags:
            data = data.with_columns(
                pl.col("nlp").map_elements(lambda x: len(
                    [token for token in x if token.pos_ == pos]),
                    return_dtype=pl.UInt16
                    ).alias(f"n_{pos.lower()}"),
            )
    elif backbone == 'stanza':
        for pos in pos_tags:
            data = data.with_columns(
                pl.col("nlp").map_elements(lambda x: len(
                    [token for sent in x.sentences for token
                     in sent.words if token.upos == pos]),
                    return_dtype=pl.UInt16
                    ).alias(f"n_{pos.lower()}"),
            )
    
    return data




"""
This module contains functions for calculating psycholinguistic features
from text data.

If you are using this module, please cite the respective sources for the
norms used in the functions. Consult them for more information on the
norms, their collection, and their interpretation in your analyses.

The norms used in this module are:

- Concreteness norms: 
    Brysbaert, M., Warriner, A. B., & Kuperman, V. (2014).
    Concreteness ratings for 40 thousand generally known English word lemmas.
    Behavior Research Methods, 46(3), 904-911.
  
- Age of acquisition norms:
    Kuperman, V., Stadthagen-Gonzalez, H., & Brysbaert, M. (2012).
    Age-of-acquisition ratings for 30,000 English words.
    Behavior Research Methods, 45(4), 1191-1207.
  
- Word prevalence norms:
    Brysbaert, M., Mandera, P., McCormick, S. F., & Keuleers, E. (2019).
    Word prevalence norms for 62,000 English lemmas.
    Behavior Research Methods, 51(2), 467-479.

- Socialness Norms:
    Diveica, V., Pexman, P.M., & Binney, R.J. (2023).
    Quantifying Social Semantics: An Inclusive Definition of Socialness
    and Ratings for 8,388 English Words.
    Behavior Research Methods, 55, 461-473.

- Iconicity norms:
    Winter, B., Lupyan, G., Perry, L. K., Dingemanse, M., & Perlman, M. (2021).
    Iconicity ratings for 14,000 English words.
    Behavior Research Methods, 56, 1640-1655.

- Sensorimotor norms:
    Lynott, D., Connell, L., Brysbaert, M., Brand, J., & Carney, J. (2020).
    The Lancaster Sensorimotor Norms: Multidimensional measures of
    perceptual and action strength for 40,000 English words.
    Behavior Research Methods, 52, 1269-1286.

The psycholinguistic features implemented in this module are:

- Abstractness/Concreteness:
    - Average concreteness score
    - Average standard deviation of the concreteness score
    - Number of low concreteness words
    - Number of high concreteness words
    - Number of controversial concreteness words

- Age of Acquisition:
    - Average age of acquisition score
    - Average standard deviation of the age of acquisition score
    - Number of low age of acquisition words
    - Number of high age of acquisition words
    - Number of controversial age of acquisition words

- Word Prevalence:
    - Average prevalence score
    - Number of low prevalence words
    - Number of high prevalence words

- Socialness:
    - Average socialness score
    - Average standard deviation of the socialness score
    - Number of low socialness words
    - Number of high socialness words
    - Number of controversial socialness words

- Iconicity:
    - Average iconicity score
    - Average standard deviation of the iconicity score
    - Number of low iconicity words
    - Number of high iconicity words
    - Number of controversial iconicity words

- Sensorimotor:
    - Average sensorimotor score
    - Average standard deviation of the sensorimotor score
    - Number of low sensorimotor words
    - Number of high sensorimotor words
    - Number of controversial sensorimotor words
"""
import polars as pl

from .preprocess import (
    get_lemmas,
)
from .util import (
    filter_lexicon,
)


# ---------------------------------------------------- #
#            Abstractness/Concreteness                 #
# ---------------------------------------------------- #

def load_concreteness_norms(path: str,
                            **kwargs: dict[str, str],
                            ) -> pl.DataFrame:
    """
    Loads the concreteness norms dataset.

    Args:
        path (str): The path to the concreteness norms dataset.

    Returns:
        concreteness_norms (pl.DataFrame):
            A Polars DataFrame containing the concreteness norms dataset.
    """
    concreteness_norms = pl.read_excel(path)

    return concreteness_norms

def filter_concreteness_norms(concreness_norms: pl.DataFrame,
                              words: list,
                              **kwargs: dict[str, str],
                              ) -> pl.DataFrame:
    """
    Filters the concreteness norms dataset by a list of words.

    Args:
        concreness_norms:
            A Polars DataFrame containing the concreteness norms.
        words (list[str]):
            A list of words to filter the concreteness norms dataset.

    Returns:
        filtered_concreteness_norms (pl.DataFrame):
            A Polars DataFrame containing the filtered concreteness norms
            dataset.
    """
    filtered_concreteness_norms = concreness_norms.filter(pl.col("Word"). \
                                                          is_in(words))
    return filtered_concreteness_norms

def get_avg_concreteness(data: pl.DataFrame,
                         lexicon: pl.DataFrame,
                         backbone: str = 'spacy',
                         **kwargs: dict[str, str],
                         ) -> pl.DataFrame:
    """
    Calculates the average concreteness score of a text. NaN/Null values
    indicate that no word in the text was found in the concreteness norms.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        lexicon (pl.DataFrame): 
            A Polars DataFrame containing the concreteness norms.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the average concreteness score
            of the text data. The average concreteness score is stored in
            a new column named 'avg_concreteness'.
    """
    if "lemmas" not in data.columns:
        data = get_lemmas(data, backbone=backbone)
    
    data = data.with_columns(
        pl.col("lemmas").map_elements(
            lambda x: filter_lexicon(lexicon=lexicon,
                                     words=x,
                                     word_column="Word"). \
                select(pl.col("Conc.M")).mean().item(),
                return_dtype=pl.Float64
                ).alias("avg_concreteness")
    )

    return data

def get_avg_sd_concreteness(data: pl.DataFrame,
                             lexicon: pl.DataFrame,
                             backbone: str = 'spacy',
                             **kwargs: dict[str, str],
                             ) -> pl.DataFrame:
    """
    Calculates the average standard deviation of concreteness score of a
    text. NaN/Null values indicate that no word in the text was found in
    the concreteness norms.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        lexicon (pl.DataFrame): 
            A Polars DataFrame containing the concreteness norms.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.
    
    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the average standard deviation
            of concreteness score of the text data.
            The average standard deviation of concreteness score is stored
            in a new column named 'avg_sd_concreteness'.
    """
    if "lemmas" not in data.columns:
        data = get_lemmas(data, backbone=backbone)

    data = data.with_columns(
        pl.col("lemmas").map_elements(
            lambda x: filter_lexicon(lexicon=lexicon,
                                     words=x,
                                     word_column="Word"). \
                select(pl.col("Conc.SD")).mean().item(),
                return_dtype=pl.Float64
                ).alias("avg_sd_concreteness")
    )

    return data

def get_n_low_concreteness(data: pl.DataFrame,
                           lexicon: pl.DataFrame,
                           threshold: float = 1.66,
                           backbone: str = 'spacy',
                           **kwargs: dict[str, str],
                           ) -> pl.DataFrame:
    """
    Calculates the number of low concreteness words in a text.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        lexicon (pl.DataFrame):
            A Polars DataFrame containing the concreteness norms.
        threshold (float): The threshold for the low concreteness words.
                    Defaults to 1.66.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the number of low concreteness
            words in the text data.
            The number of low concreteness words is stored in a new column
            named 'n_low_concreteness'.
    """
    if "lemmas" not in data.columns:
        data = get_lemmas(data, backbone=backbone)
    data = data.with_columns(
        pl.col("lemmas").map_elements(
            lambda x: filter_lexicon(lexicon=lexicon,
                                     words=x,
                                     word_column="Word"). \
                select(pl.col("Conc.M")).filter(
                    pl.col("Conc.M") < threshold). \
                    count().item(),
                return_dtype=pl.Int64).alias("n_low_concreteness")
    )

    return data

def get_n_high_concreteness(data: pl.DataFrame,
                            lexicon: pl.DataFrame,
                            backbone: str = 'spacy',
                            threshold: float = 3.33,
                            **kwargs: dict[str, str],
                            ) -> pl.DataFrame:
    """
    Calculates the number of high concreteness words in a text.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        lexicon (pl.DataFrame):
            A Polars DataFrame containing the concreteness norms.
        threshold (float): The threshold for the high concreteness words.
                    Defaults to 3.33.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the number of high concreteness
            words in the text data.
            The number of high concreteness words is stored in a new
            column named 'n_high_concreteness'.
    """
    if "lemmas" not in data.columns:
        data = get_lemmas(data, backbone=backbone)
    data = data.with_columns(
        pl.col("lemmas").map_elements(
            lambda x: filter_lexicon(lexicon=lexicon,
                                     words=x,
                                     word_column="Word"). \
                select(pl.col("Conc.M")).filter(
                    pl.col("Conc.M") > threshold). \
                    count().item(),
                return_dtype=pl.Int64).alias("n_high_concreteness")
    )

    return data

def get_n_controversial_concreteness(data: pl.DataFrame,
                                     lexicon: pl.DataFrame,
                                     backbone: str = 'spacy',
                                     threshold: float = 2.0,
                                     **kwargs: dict[str, str],
                                     ) -> pl.DataFrame:
    """
    Calculates the number of controversial concreteness words in a text
    (i.e. items with a high standard deviation in the ratings).

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        lexicon (pl.DataFrame):
            A Polars DataFrame containing the concreteness norms.
        threshold (float): The threshold for the standard deviation.
                 Defaults to 2.5.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the number of controversial
            concreteness words in the text data.
            The number of controversial concreteness words is stored in a
            new column named 'n_controversial_concreteness'.
    """
    if "lemmas" not in data.columns:
        data = get_lemmas(data, backbone=backbone)
    
    data = data.with_columns(
        pl.col("lemmas").map_elements(
            lambda x: filter_lexicon(lexicon=lexicon,
                                     words=x,
                                     word_column="Word"). \
                select(pl.col("Conc.SD")).filter(
                    pl.col("Conc.SD") > threshold). \
                    count().item(),
                return_dtype=pl.Int64).alias("n_controversial_concreteness")
    )

    return data

# ---------------------------------------------------- #
#            Age of Acquisition                        #
# ---------------------------------------------------- #

def load_aoa_norms(path: str,
                   **kwargs: dict[str, str],
                   ) -> pl.DataFrame:
    """
    Loads the age of acquisition norms dataset.

    Args:
        path (str): The path to the age of acquisition norms dataset.

    Returns:
        aoa_norms (pl.DataFrame):
            A Polars DataFrame containing the age of acquisition norms
            dataset.
    """
    aoa_norms = pl.read_excel(path)
    
    return aoa_norms

def get_avg_aoa(data: pl.DataFrame,
                lexicon: pl.DataFrame,
                backbone: str = 'spacy',
                **kwargs: dict[str, str],
                ) -> pl.DataFrame:
    """
    Calculates the average age of acquisition score of a text. NaN/Null
    values indicate that no word in the text was found in the age of
    acquisition norms.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        lexicon (pl.DataFrame):
            A Polars DataFrame containing the age of acquisition norms.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the average age of acquisition
            score of the text data.
            The average age of acquisition score is stored in a new column
            named 'avg_aoa'.
    """
    if "lemmas" not in data.columns:
        data = get_lemmas(data, backbone=backbone)
    data = data.with_columns(
        pl.col("lemmas").map_elements(
            lambda x: filter_lexicon(lexicon=lexicon,
                                     words=x,
                                     word_column="Word"). \
                select(pl.col("Rating.Mean")).mean().item(),
                return_dtype=pl.Float64
                ).alias("avg_aoa")
    )

    return data

def get_avg_sd_aoa(data: pl.DataFrame,
                    lexicon: pl.DataFrame,
                    backbone: str = 'spacy',
                    **kwargs: dict[str, str],
                    ) -> pl.DataFrame:
    """
    Calculates the average standard deviation of age of acquisition score 
    of a text. NaN/Null values indicate that no word in the text was found
    in the age of acquisition norms.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        lexicon (pl.DataFrame):
            A Polars DataFrame containing the age of acquisition norms.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.
    
    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the average standard deviation
            of age of acquisition score of the text data.
            The average standard deviation of age of acquisition score is 
            stored in a new column named 'avg_sd_aoa'.
    """
    if "lemmas" not in data.columns:
        data = get_lemmas(data, backbone=backbone)

    data = data.with_columns(
        pl.col("lemmas").map_elements(
            lambda x: filter_lexicon(lexicon=lexicon,
                                     words=x,
                                     word_column="Word"). \
                select(pl.col("Rating.SD")).mean().item(),
                return_dtype=pl.Float64
                ).alias("avg_sd_aoa")
    )

    return data

def get_n_low_aoa(data: pl.DataFrame,
                  lexicon: pl.DataFrame,
                  backbone: str = 'spacy',
                  threshold: float = 10.0,
                  **kwargs: dict[str, str],
                  ) -> pl.DataFrame:
        """
        Calculates the number of low age of acquisition words in a text.

        Args:
            data (pl.DataFrame):
                A Polars DataFrame containing the text data.
            aoa_norms (pl.DataFrame): 
                A Polars DataFrame containing the age of acquisition norms.
            backbone (str): The NLP library used to process the text data.
                    Either 'spacy' or 'stanza'. 
            threshold (float): The threshold for the low age of acquisition words.
                    Defaults to 10.0.

        Returns:
            data (pl.DataFrame):
                A Polars DataFrame containing the number of low age of
                acquisition words in the text data.
                The number of low age of acquisition words is stored in a
                new column named 'n_low_aoa'.
        """
        if "lemmas" not in data.columns:
            data = get_lemmas(data, backbone=backbone)
        data = data.with_columns(
            pl.col("lemmas").map_elements(
                lambda x: filter_lexicon(lexicon=lexicon,
                                         words=x,
                                         word_column="Word"). \
                    select(pl.col("Rating.Mean")
                           ).filter(pl.col("Rating.Mean") < threshold). \
                        count().item(),
                    return_dtype=pl.Int64).alias("n_low_aoa")
        )
    
        return data

def get_n_high_aoa(data: pl.DataFrame,
                   lexicon: pl.DataFrame,
                   threshold: float = 10.0,
                   backbone: str = 'spacy',
                   **kwargs: dict[str, str],
                   ) -> pl.DataFrame:
        """
        Calculates the number of high age of acquisition words in a text.

        Args:
            data (pl.DataFrame): A Polars DataFrame containing the text data.
            lexicon (pl.DataFrame):
                A Polars DataFrame containing the age of acquisition norms.
            threshold (float): The threshold for the high age of acquisition words.
                    Defaults to 10.0.

        Returns:
            data (pl.DataFrame):
                A Polars DataFrame containing the number of high age of
                acquisition words in the text data.
                The number of high age of acquisition words is stored in a new
                column named 'n_high_aoa'.
        """
        if "lemmas" not in data.columns:
            data = get_lemmas(data, backbone=backbone)
        data = data.with_columns(
            pl.col("lemmas").map_elements(
                lambda x: filter_lexicon(lexicon=lexicon,
                                         words=x,
                                         word_column="Word"). \
                    select(pl.col("Rating.Mean")).filter(
                        pl.col("Rating.Mean") > threshold
                        ).count().item(),
                    return_dtype=pl.Int64).alias("n_high_aoa")
        )
    
        return data

def get_n_controversial_aoa(data: pl.DataFrame,
                            lexicon: pl.DataFrame,
                            backbone: str = 'spacy',
                            threshold: float = 4.5,
                            **kwargs: dict[str, str],
                            ) -> pl.DataFrame:
    """
    Calculates the number of controversial age of acquisition words in a
    text (i.e. items with a high standard deviation in the ratings).

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        lexicon (pl.DataFrame):
            A Polars DataFrame containing the age of acquisition norms.
        threshold (float): The threshold for the standard deviation.
                 Defaults to 4.5.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.
    
    Returns:
        data (pl.DataFrame): 
            A Polars DataFrame containing the number of controversial age
            of acquisition words in the text data.
            The number of controversial age of acquisition words is stored
            in a new column named 'n_controversial_aoa'.
    """
    if "lemmas" not in data.columns:
        data = get_lemmas(data, backbone=backbone)

    data = data.with_columns(
        pl.col("lemmas").map_elements(
            lambda x: filter_lexicon(lexicon=lexicon,
                                     words=x,
                                     word_column="Word"). \
                select(pl.col("Rating.SD")).filter(
                    pl.col("Rating.SD") > threshold). \
                    count().item(),
                return_dtype=pl.Int64).alias("n_controversial_aoa")
    )

    return data

# ---------------------------------------------------- #
#                Word Prevalence                       #
# ---------------------------------------------------- #

# Note: Prevalence norms used here do not have a standard deviation column.

def load_prevalence_norms(path: str,
                          **kwargs: dict[str, str],
                          ) -> pl.DataFrame:
    """
    Loads the word prevalence norms dataset.

    Args:
        path (str): The path to the word prevalence norms dataset.

    Returns:
        prevalence_norms (pl.DataFrame):
            A Polars DataFrame containing the word prevalence norms
            dataset.
    """
    prevalence_norms = pl.read_excel(path)

    return prevalence_norms

def get_avg_prevalence(data: pl.DataFrame,
                       lexicon: pl.DataFrame,
                       backbone: str = 'spacy',
                       **kwargs: dict[str, str],
                       ) -> pl.DataFrame:
    """
    Calculates the average prevalence score of a text. NaN/Null values
    indicate that no word in the text was found in the prevalence norms.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        prevalence_norms (pl.DataFrame):
            A Polars DataFrame containing the word prevalence norms.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the average prevalence score
            of the text data. The average prevalence score is stored in a 
            new column named 'avg_prevalence'.
    """
    if "lemmas" not in data.columns:
        data = get_lemmas(data, backbone=backbone)
    data = data.with_columns(
        pl.col("lemmas").map_elements(
            lambda x: filter_lexicon(lexicon=lexicon,
                                     words=x,
                                     word_column="Word"). \
                select(pl.col("Prevalence")).mean().item(),
                return_dtype=pl.Float64
                ).alias("avg_prevalence")
    )

    return data              

def get_n_low_prevalence(data: pl.DataFrame,
                         lexicon: pl.DataFrame,
                         threshold: float = 1.0,
                         backbone: str = 'spacy',
                         **kwargs: dict[str, str],
                         ) -> pl.DataFrame:
    """
    Calculate the number of low prevalence words in a text.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        lexicon (pl.DataFrame):
            A Polars DataFrame containing the word prevalence norms.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the number of low prevalence
            words in the text data. The number of low prevalence words is 
            stored in a new column named 'n_low_prevalence'.
    """
    if "lemmas" not in data.columns:
        data = get_lemmas(data, backbone=backbone)
    data = data.with_columns(
        pl.col("lemmas").map_elements(
            lambda x: filter_lexicon(lexicon=lexicon,
                                     words=x,
                                     word_column="Word"). \
                select(pl.col("Prevalence")).filter(
                    pl.col("Prevalence") < threshold). \
                    count().item(),
                return_dtype=pl.Int64).alias("n_low_prevalence")
    )

    return data

def get_n_high_prevalence(data: pl.DataFrame,
                          lexicon: pl.DataFrame,
                          backbone: str = 'spacy',
                          threshold: float = 1.0,
                          **kwargs: dict[str, str],
                          ) -> pl.DataFrame:
    """
    Calculate the number of high prevalence words in a text.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        lexicons: A Polars DataFrame containing the word prevalence norms.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.
        threshold (float): The threshold for the high prevalence words.
                    Defaults to 1.0.

    Returns:
        data (pl.DataFrame): 
            A Polars DataFrame containing the number of high prevalence
            words in the text data.
            The number of high prevalence words is stored in a new column 
            named 'n_high_prevalence'.
    """
    if "lemmas" not in data.columns:
        data = get_lemmas(data, backbone=backbone)
    data = data.with_columns(
        pl.col("lemmas").map_elements(
            lambda x: filter_lexicon(lexicon=lexicon,
                                     words=x,
                                     word_column="Word"). \
                select(pl.col("Prevalence")).filter(
                    pl.col("Prevalence") > threshold). \
                    count().item(),
                return_dtype=pl.Int64).alias("n_high_prevalence")
    )

    return data

# ---------------------------------------------------- #
#                Socialness Norms                      #
# ---------------------------------------------------- #

def load_socialness_norms(path: str,
                          **kwargs: dict[str, str],
                          ) -> pl.DataFrame:
    """
    Loads the socialness norms dataset.

    Args:
        path (str): The path to the socialness norms dataset.

    Returns:
        socialness_norms (pl.DataFrame): 
            A Polars DataFrame containing the socialness norms dataset.
    """
    socialness_norms = pl.read_csv(path)

    return socialness_norms

def get_avg_socialness(data: pl.DataFrame,
                       lexicon: pl.DataFrame,
                       backbone: str = 'spacy',
                       **kwargs: dict[str, str],
                       ) -> pl.DataFrame:
    """
    Calculates the average socialness score of a text. NaN/Null values
    indicate that no word in the text was found in the socialness norms.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        lexicon (pl.DataFrame):
            A Polars DataFrame containing the socialness norms.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the average socialness score
            of the text data. The average socialness score is stored in a 
            new column named 'avg_socialness'.
    """
    if "lemmas" not in data.columns:
        data = get_lemmas(data, backbone=backbone)
    data = data.with_columns(
        pl.col("lemmas").map_elements(
            lambda x: filter_lexicon(lexicon=lexicon,
                                     words=x,
                                     word_column="Word"). \
                select(pl.col("Mean")).mean().item(),
                return_dtype=pl.Float64
                ).alias("avg_socialness")
    )

    return data

def get_avg_sd_socialness(data: pl.DataFrame,
                          lexicon: pl.DataFrame,
                          backbone: str = 'spacy',
                          **kwargs: dict[str, str],
                          ) -> pl.DataFrame:
    """
    Calculates the average standard deviation of socialness score of a
    text. NaN/Null values indicate that no word in the text was found in
    the socialness norms.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        lexicon (pl.DataFrame):
            A Polars DataFrame containing the socialness norms.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the average standard deviation
            of socialness score of the text data.
            The average standard deviation of socialness score is stored
            in a new column named 'avg_sd_socialness'.
    """
    if "lemmas" not in data.columns:
        data = get_lemmas(data, backbone=backbone)

    data = data.with_columns(
        pl.col("lemmas").map_elements(
            lambda x: filter_lexicon(lexicon=lexicon,
                                     words=x,
                                     word_column="Word"). \
                select(pl.col("SD")).mean().item(),
                return_dtype=pl.Float64
                ).alias("avg_sd_socialness")
    )

    return data

def get_n_low_socialness(data: pl.DataFrame,
                         lexicon: pl.DataFrame,
                         threshold: float = 2.33,
                         backbone: str = 'spacy',
                         **kwargs: dict[str, str],
                         ) -> pl.DataFrame:
    """
    Calculates the number of low socialness words in a text.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        lexicon (pl.DataFrame):
            A Polars DataFrame containing the socialness norms.
        threshold (float): The threshold for the low socialness words.
                    Defaults to 1.66.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the number of low socialness
            words in the text data. The number of low socialness words is 
            stored in a new column named 'n_low_socialness'.
    """
    if "lemmas" not in data.columns:
        data = get_lemmas(data, backbone=backbone)
    data = data.with_columns(
        pl.col("lemmas").map_elements(
            lambda x: filter_lexicon(lexicon=lexicon,
                                     words=x,
                                     word_column="Word"). \
                select(pl.col("Mean")).filter(
                    pl.col("Mean") < threshold). \
                    count().item(),
                return_dtype=pl.Int64).alias("n_low_socialness")
    )

    return data

def get_n_high_socialness(data: pl.DataFrame,
                          lexicon: pl.DataFrame,
                          backbone: str = 'spacy',
                          threshold: float = 3.66,
                          **kwargs: dict[str, str],
                          ) -> pl.DataFrame:
    """
    Calculates the number of high socialness words in a text.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        lexicon (pl.DataFrame):
            A Polars DataFrame containing the socialness norms.
        threshold (float): The threshold for the high socialness words.
                    Defaults to 3.66.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.
    
    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the number of high socialness
            words in the text data. The number of high socialness words is
            stored in a new column named 'n_high_socialness'.
    """
    if "lemmas" not in data.columns:
        data = get_lemmas(data, backbone=backbone)
    data = data.with_columns(
        pl.col("lemmas").map_elements(
            lambda x: filter_lexicon(lexicon=lexicon,
                                     words=x,
                                     word_column="Word"). \
                select(pl.col("Mean")).filter(
                    pl.col("Mean") > threshold). \
                    count().item(),
                return_dtype=pl.Int64).alias("n_high_socialness")
    )

    return data

def get_n_controversial_socialness(data: pl.DataFrame,
                                   lexicon: pl.DataFrame,
                                   backbone: str = 'spacy',
                                   threshold: float = 2.0,
                                   **kwargs: dict[str, str],
                                   ) -> pl.DataFrame:
    """
    Calculates the number of controversial socialness words in a text
    (i.e. items with a high standard deviation in the ratings).

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        lexicon (pl.DataFrame): 
            A Polars DataFrame containing the socialness norms.
        threshold (float): The threshold for the standard deviation.
                 Defaults to 2.0.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.
    
    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the number of controversial
            socialness words in the text data.
            The number of controversial socialness words is stored in a
            new column named 'n_controversial_socialness'.
    """
    if "lemmas" not in data.columns:
        data = get_lemmas(data, backbone=backbone)

    data = data.with_columns(
        pl.col("lemmas").map_elements(
            lambda x: filter_lexicon(lexicon=lexicon,
                                     words=x,
                                     word_column="Word"). \
                select(pl.col("SD")).filter(pl.col("SD") > threshold). \
                    count().item(),
                return_dtype=pl.Int64).alias("n_controversial_socialness")
    )

    return data

# ---------------------------------------------------- #
#                      Iconicity                       #
# ---------------------------------------------------- #

def load_iconicity_norms(path: str,
                         **kwargs: dict[str, str],
                         ) -> pl.DataFrame:
    """
    Loads the iconicity norms dataset.

    Args:
        path (str): The path to the iconicity norms dataset.

    Returns:
        iconicity_norms (pl.DataFrame):
            A Polars DataFrame containing the iconicity norms dataset.
    """
    iconicity_norms = pl.read_csv(path)
    return iconicity_norms

def get_avg_iconicity(data: pl.DataFrame,
                        lexicon: pl.DataFrame,
                        backbone: str = 'spacy',
                        **kwargs: dict[str, str],
                        ) -> pl.DataFrame:
    """
    Calculates the average iconicity score of a text. NaN/Null values
    indicate that no word in the text was found in the iconicity norms.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        lexicon (pl.DataFrame):
            A Polars DataFrame containing the iconicity norms.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the average iconicity score
            of the text data. The average iconicity score is stored in a
            new column named 'avg_iconicity'.
    """
    if "lemmas" not in data.columns:
        data = get_lemmas(data, backbone=backbone)
    data = data.with_columns(
        pl.col("lemmas").map_elements(
            lambda x: filter_lexicon(lexicon=lexicon,
                                     words=x,
                                     word_column="word"). \
                select(pl.col("rating")).mean().item(),
                return_dtype=pl.Float64
                ).alias("avg_iconicity")
    )

    return data

def get_avg_sd_iconicity(data: pl.DataFrame,
                          lexicon: pl.DataFrame,
                          backbone: str = 'spacy',
                          **kwargs: dict[str, str],
                          ) -> pl.DataFrame:
    """
    Calculates the average standard deviation of iconicity score of a text.
    NaN/Null values indicate that no word in the text was found in the
    iconicity norms.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        lexicon (pl.DataFrame):
            A Polars DataFrame containing the iconicity norms.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the average standard deviation
            of iconicity score of the text data.
            The average standard deviation of iconicity score is stored in
            a new column named 'avg_sd_iconicity'.
    """
    if "lemmas" not in data.columns:
        data = get_lemmas(data, backbone=backbone)

    data = data.with_columns(
        pl.col("lemmas").map_elements(
            lambda x: filter_lexicon(lexicon=lexicon,
                                     words=x,
                                     word_column="word"). \
                select(pl.col("rating_sd")).mean().item(),
                return_dtype=pl.Float64
                ).alias("avg_sd_iconicity")
    )

    return data

def get_n_low_iconicity(data: pl.DataFrame,
                        lexicon: pl.DataFrame,
                        threshold: float = 2.33,
                        backbone: str = 'spacy',
                        **kwargs: dict[str, str],
                        ) -> pl.DataFrame:
    """
    Calculates the number of low iconicity words in a text.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        lexicon (pl.DataFrame):
            A Polars DataFrame containing the iconicity norms.
        threshold (float): The threshold for the low iconicity words.
                    Defaults to 2.33.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the number of low iconicity
            words in the text data. The number of low iconicity words is
            stored in a new column named 'n_low_iconicity'.
    """
    if "lemmas" not in data.columns:
        data = get_lemmas(data, backbone=backbone)
    data = data.with_columns(
        pl.col("lemmas").map_elements(
            lambda x: filter_lexicon(lexicon=lexicon,
                                     words=x,
                                     word_column="word"). \
                select(pl.col("rating")).filter(
                    pl.col("rating") < threshold). \
                    count().item(),
                return_dtype=pl.Int64).alias("n_low_iconicity")
    )

    return data

def get_n_high_iconicity(data: pl.DataFrame,
                         lexicon: pl.DataFrame,
                         backbone: str = 'spacy',
                         threshold: float = 3.66,
                         **kwargs: dict[str, str],
                         ) -> pl.DataFrame:
    """
    Calculates the number of high iconicity words in a text.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        lexicon (pl.DataFrame):
            A Polars DataFrame containing the iconicity norms.
        threshold (float): The threshold for the high iconicity words.
                    Defaults to 3.66.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the number of high iconicity
            words in the text data.
            The number of high iconicity words is stored in a new column
            named 'n_high_iconicity'.
    """
    if "lemmas" not in data.columns:
        data = get_lemmas(data, backbone=backbone)
    data = data.with_columns(
        pl.col("lemmas").map_elements(
            lambda x: filter_lexicon(lexicon=lexicon,
                                     words=x,
                                     word_column="word"). \
                select(pl.col("rating")).filter(
                    pl.col("rating") > threshold). \
                    count().item(),
                return_dtype=pl.Int64).alias("n_high_iconicity")
    )

    return data

def get_n_controversial_iconicity(data: pl.DataFrame,
                                  lexicon: pl.DataFrame,
                                  backbone: str = 'spacy',
                                  threshold: float = 2.5,
                                  **kwargs: dict[str, str],
                                  ) -> pl.DataFrame:
    """
    Calculates the number of controversial iconicity words in a text
    (i.e. items with a high standard deviation in the ratings).

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        lexicon (pl.DataFrame):
            A Polars DataFrame containing the iconicity norms.
        threshold (float): The threshold for the standard deviation.
                 Defaults to 2.5.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.
    
    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the number of controversial
            iconicity words in the text data.
            The number of controversial iconicity words is stored in a new
            column named 'n_controversial_iconicity'.
    """
    if "lemmas" not in data.columns:
        data = get_lemmas(data, backbone=backbone)

    data = data.with_columns(
        pl.col("lemmas").map_elements(
            lambda x: filter_lexicon(lexicon=lexicon,
                                     words=x,
                                     word_column="word"). \
                select(pl.col("rating_sd")).filter(
                    pl.col("rating_sd") > threshold). \
                    count().item(),
                return_dtype=pl.Int64).alias("n_controversial_iconicity")
    )

    return data

# ---------------------------------------------------- #
#                   Sensorimotor                       #
# ---------------------------------------------------- #

SENSORIMOTOR_VARS = [
    "Auditory",
    "Gustatory",
    "Haptic",
    "Interoceptive",
    "Olfactory",
    "Visual",
    "Foot_leg",
    "Hand_arm",
    "Head",
    "Mouth",
    "Torso",
]

def load_sensorimotor_norms(path: str,
                            **kwargs: dict[str, str],
                            ) -> pl.DataFrame:
    """
    Loads the sensorimotor norms dataset.

    Args:
        path (str): The path to the sensorimotor norms dataset.

    Returns:
        sensorimotor_norms (pl.DataFrame):
            A Polars DataFrame containing the sensorimotor norms dataset.
    """
    sensorimotor_norms = pl.read_excel(path)
    # the lancaster sensorimotor norms have all caps lemmas, so we con-
    # vert them to lowercase for consistency
    sensorimotor_norms = sensorimotor_norms.with_columns(
        pl.col("Word").str.to_lowercase().alias("Word")
    )

    return sensorimotor_norms

def get_avg_sensorimotor(data: pl.DataFrame,
                         lexicon: pl.DataFrame,
                         backbone: str = 'spacy',
                         sensorimotor_vars: list[str] = SENSORIMOTOR_VARS,
                         **kwargs: dict[str, str],
                         ) -> pl.DataFrame:
    """
    Calculates the average sensorimotor variable score of a text.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        lexicon (pl.DataFrame):
            A Polars DataFrame containing the sensorimotor norms.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.
        sensorimotor_vars (list[str]):
            A list of sensorimotor variables to calculate the average
            score for. Defaults to SENSORIMOTOR_VARS.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the average sensorimotor score
            of the text data.
            The average sensorimotor score is stored in a new column named
            'avg_sensorimotor'.
    """
    if "lemmas" not in data.columns:
        data = get_lemmas(data, backbone=backbone)
    for var in sensorimotor_vars:
        data = data.with_columns(
            pl.col("lemmas").map_elements(
                lambda x: filter_lexicon(lexicon=lexicon,
                                         words=x,
                                         word_column="Word"). \
                    select(pl.col(f"{var}.mean")).mean().item(),
                    return_dtype=pl.Float64
                    ).alias(f"avg_{var}_sensorimotor")
        )

    return data

def get_avg_sd_sensorimotor(data: pl.DataFrame,
                            lexicon: pl.DataFrame,
                            backbone: str = 'spacy',
                            sensorimotor_vars: list[str] = \
                                SENSORIMOTOR_VARS,
                            **kwargs: dict[str, str],
                            ) -> pl.DataFrame:
    """
    Calculates the average standard deviation of sensorimotor variable
    score of a text.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        lexicon (pl.DataFrame):
            A Polars DataFrame containing the sensorimotor norms.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.
        sensorimotor_vars (list[str]):
            A list of sensorimotor variables to calculate the average
            standard deviation for.
    
    Returns:
        data (pl.DataFrame): A Polars DataFrame containing the average
            standard deviation of sensorimotor variable score of the text
            data. The average standard deviation of sensorimotor variable 
            score is stored in new columns named 'avg_sd_{var}' where
            {var} is the sensorimotor variable.
    """
    if "lemmas" not in data.columns:
        data = get_lemmas(data, backbone=backbone)
    
    for var in sensorimotor_vars:
        data = data.with_columns(
            pl.col("lemmas").map_elements(
                lambda x: filter_lexicon(lexicon=lexicon,
                                         words=x,
                                         word_column="Word"). \
                    select(pl.col(f"{var}.SD")).mean().item(),
                    return_dtype=pl.Float64
                    ).alias(f"avg_sd_{var}_sensorimotor")
        )

    return data

def get_n_low_sensorimotor(data: pl.DataFrame,
                           lexicon: pl.DataFrame,
                           threshold: float = 2.33,
                           backbone: str = 'spacy',
                           sensorimotor_vars: list[str] = \
                            SENSORIMOTOR_VARS,
                           **kwargs: dict[str, str],
                           ) -> pl.DataFrame:
    """
    Calculates the number of low-rating sensorimotor words in a text.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        lexicon (pl.DataFrame):
            A Polars DataFrame containing the sensorimotor norms.
        threshold (float): The threshold for the low-rating sensorimotor words.
                    Defaults to 2.33.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.
        sensorimotor_vars (list[str]):
            A list of sensorimotor variables to calculate the number of
            low-rating words for. Defaults to SENSORIMOTOR_VARS.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the number of low-rating
            sensorimotor words in the text data.
            The number of low-rating sensorimotor words is stored in new
            columns named 'n_low_{var}' where {var} is the sensorimotor
            variable.
    """
    if "lemmas" not in data.columns:
        data = get_lemmas(data, backbone=backbone)
    
    for var in sensorimotor_vars:
        data = data.with_columns(
            pl.col("lemmas").map_elements(
                lambda x: filter_lexicon(lexicon=lexicon,
                                         words=x,
                                         word_column="Word"). \
                    select(pl.col(f"{var}.mean")).filter(
                        pl.col(f"{var}.mean") < threshold). \
                        count().item(),
                    return_dtype=pl.Int64).alias(f"n_low_{var}"
                                                 "_sensorimotor")
        )

    return data

def get_n_high_sensorimotor(data: pl.DataFrame,
                            lexicon: pl.DataFrame,
                            backbone: str = 'spacy',
                            threshold: float = 3.66,
                            sensorimotor_vars: list[str] = \
                                SENSORIMOTOR_VARS,
                            **kwargs: dict[str, str],
                            ) -> pl.DataFrame:
    """
    Calculates the number of high-rating sensorimotor words in a text.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        lexicon (pl.DataFrame):
            A Polars DataFrame containing the sensorimotor norms.
        threshold (float): The threshold for the high-rating sensorimotor words.
                    Defaults to 3.66.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.
        sensorimotor_vars (list[str]):
            A list of sensorimotor variables to calculate the
            number of high-rating words for.
            Defaults to SENSORIMOTOR_VARS.
    
    Returns:
        data (pl.DataFrame): 
            A Polars DataFrame containing the number of high-rating
            sensorimotor words in the text data.
            The number of high-rating sensorimotor words is stored in new 
            columns named 'n_high_{var}' where {var} is the sensorimotor
            variable.
    """
    if "lemmas" not in data.columns:
        data = get_lemmas(data, backbone=backbone)
    
    for var in sensorimotor_vars:
        data = data.with_columns(
            pl.col("lemmas").map_elements(
                lambda x: filter_lexicon(lexicon=lexicon,
                                         words=x,
                                         word_column="Word"). \
                    select(pl.col(f"{var}.mean")).filter(
                        pl.col(f"{var}.mean") > threshold). \
                        count().item(),
                    return_dtype=pl.Int64).alias(f"n_high_{var}"
                                                 "_sensorimotor")
        )

    return data

def get_n_controversial_sensorimotor(data: pl.DataFrame,
                                     lexicon: pl.DataFrame,
                                     backbone: str = 'spacy',
                                     threshold: float = 2.0,
                                     sensorimotor_vars: list[str] = \
                                        SENSORIMOTOR_VARS,
                                     **kwargs: dict[str, str],
                                     ) -> pl.DataFrame:
    """
    Calculates the number of controversial sensorimotor words in a text
    (i.e. items with a high standard deviation in the ratings).

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        lexicon (pl.DataFrame): 
            A Polars DataFrame containing the sensorimotor norms.
        threshold (float): The threshold for the standard deviation.
                 Defaults to 2.0.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.
        sensorimotor_vars (list[str]):
            A list of sensorimotor variables to calculate the
            number of controversial words for.
    
    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the number of controversial
            sensorimotor words in the text data.
            The number of controversial sensorimotor words is stored in
            new columns named 'n_controversial_{var}' where {var} is the
            sensorimotor variable.
    """
    if "lemmas" not in data.columns:
        data = get_lemmas(data, backbone=backbone)

    for var in sensorimotor_vars:
        data = data.with_columns(
            pl.col("lemmas").map_elements(
                lambda x: filter_lexicon(lexicon=lexicon,
                                         words=x,
                                         word_column="Word"). \
                    select(pl.col(f"{var}.SD")).filter(
                        pl.col(f"{var}.SD") > threshold). \
                        count().item(),
                    return_dtype=pl.Int64).alias(f"n_controversial_{var}"
                                                 "_sensorimotor")
        )

    return data

